import os
from typing import overload

from anvil import ANVIL, CONFIG
from anvil.api.actors._animations import _BPAnimations
from anvil.api.actors._component_group import (_ComponentGroup, _Components,
                                               _Properties)
from anvil.api.actors._events import _Event
from anvil.api.actors._render_controller import _RenderControllers
from anvil.api.actors.components import EntityInstantDespawn, EntityRideable
from anvil.api.actors.spawn_rules import SpawnRule
from anvil.api.logic.molang import Molang, Variable
from anvil.api.pbr.pbr import __TextureSet
from anvil.api.vanilla.entities import MinecraftEntityTypes
from anvil.api.vanilla.items import MinecraftItemTypes
from anvil.lib.blockbench import _Blockbench
from anvil.lib.config import ConfigPackageTarget
from anvil.lib.enums import DamageSensor, Target
from anvil.lib.lib import MOLANG_PREFIXES
from anvil.lib.reports import ReportType
from anvil.lib.schemas import (AddonObject, EntityDescriptor, JsonSchemes,
                               MinecraftDescription, MinecraftEntityDescriptor)
from anvil.lib.sounds import EntitySoundEvent, SoundCategory, SoundDescription
from anvil.lib.types import RGB, RGBA, Vector2D

__all__ = ["Entity", "Attachable"]


class _BP_ControllerState:
    def __init__(self, state_name):
        self._state_name = state_name
        self._controller_state: dict = JsonSchemes.animation_controller_state(self._state_name)
        self._default = True

    def on_entry(self, *commands: str):
        """Events, commands or molang to preform on entry of this state.

        Args:
            *commands (str): The Events, commands or molang to preform on entry of this state.

        Returns:
            This state.
        """
        for command in commands:
            if str(command).startswith(Target.S):
                self._controller_state[self._state_name]["on_entry"].append(command)
            elif any(str(command).startswith(v) for v in MOLANG_PREFIXES):
                self._controller_state[self._state_name]["on_entry"].append(f"{command};")
            else:
                self._controller_state[self._state_name]["on_entry"].append(f"/{command}")
        self._default = False
        return self

    def on_exit(self, *commands: str):
        """Events, commands or molang to preform on exit of this state.

        Args:
            *commands (str): The Events, commands or molang to preform on exit of this state.

        Returns:
            This state.
        """
        for command in commands:
            if str(command).startswith(Target.S.value):
                self._controller_state[self._state_name]["on_exit"].append(f"{command}")
            elif any(str(command).startswith(v) for v in MOLANG_PREFIXES):
                self._controller_state[self._state_name]["on_exit"].append(f"{command};")
            else:
                self._controller_state[self._state_name]["on_exit"].append(f"/{command}")
        self._default = False
        return self

    def animation(self, animation: str, condition: str = None):
        """Animation short name to play during this state.

        Args:
            animation (str): The name of the animation to play.
            condition (str, optional): The condition on which this animation plays.

        Returns:
            This state.
        """
        if condition is None:
            self._controller_state[self._state_name]["animations"].append(animation)
        else:
            self._controller_state[self._state_name]["animations"].append({animation: condition})
        self._default = False
        return self

    def transition(self, state: str, condition: str):
        """Target state to switch to and the condition to do so.

        Parameters
        ----------
        state : str
            The name of the state to transition to.
        condition : str
            The condition that must be met for the transition to occur.

        Returns
        -------
            This state.

        """
        self._controller_state[self._state_name]["transitions"].append({state: str(condition)})
        self._default = False
        return self

    @property
    def _export(self):
        return self._controller_state


class _RP_ControllerState:
    def __init__(self, actor: "_ActorClientDescription", state_name):
        self._actor = actor
        self._state_name = state_name
        self._controller_state = JsonSchemes.animation_controller_state(self._state_name)
        self._controller_state[self._state_name]["particle_effects"] = []
        self._controller_state[self._state_name]["sound_effects"] = []
        self._default = True

    @overload
    def on_entry(self, *commands: str | Molang) -> "_RP_ControllerState": ...

    @overload
    def on_entry(self, variable: Molang | str, value: Molang | str | int | float) -> "_RP_ControllerState": ...

    def on_entry(self, *args, **kwargs) -> "_RP_ControllerState":
        """molang to preform on entry of this state.

        Parameters
        ----------
        *args : str | Molang or (variable, value)
            param args: molang commands to perform on entry of this state, or variable assignment.

        Returns
        -------
            This state.

        """
        # Handle variable assignment case: on_entry(variable, value)
        if len(args) == 2 and not kwargs:
            variable, value = args
            command = f"{variable}={value}"
            if any(command.startswith(v) for v in MOLANG_PREFIXES):
                self._controller_state[self._state_name]["on_entry"].append(f"{command};")
            else:
                raise RuntimeError(
                    f"Invalid variable assignment for on_entry: {command}. Only Molang commands are allowed. Resource Pack Controller State [{self._state_name}]"
                )
        # Handle commands case: on_entry(*commands)
        else:
            for command in args:
                if any(str(command).startswith(v) for v in MOLANG_PREFIXES):
                    self._controller_state[self._state_name]["on_entry"].append(f"{command};")
                else:
                    raise RuntimeError(
                        f"Invalid command for on_entry: {command}. Only Molang commands are allowed. Resource Pack Controller State [{self._state_name}]"
                    )

        self._default = False
        return self

    @overload
    def on_exit(self, *commands: str | Molang) -> "_RP_ControllerState": ...

    @overload
    def on_exit(self, variable: Molang | str, value: Molang | str | int | float) -> "_RP_ControllerState": ...

    def on_exit(self, *args, **kwargs) -> "_RP_ControllerState":
        """molang to preform on entry of this state.

        Parameters
        ----------
        *args : str | Molang or (variable, value)
            param args: molang commands to perform on entry of this state, or variable assignment.

        Returns
        -------
            This state.

        """
        # Handle variable assignment case: on_exit(variable, value)
        if len(args) == 2 and not kwargs:
            variable, value = args
            command = f"{variable}={value}"
            if any(command.startswith(v) for v in MOLANG_PREFIXES):
                self._controller_state[self._state_name]["on_exit"].append(f"{command};")
            else:
                raise RuntimeError(
                    f"Invalid variable assignment for on_exit: {command}. Only Molang commands are allowed. Resource Pack Controller State [{self._state_name}]"
                )
        # Handle commands case: on_exit(*commands)
        else:
            for command in args:
                if any(str(command).startswith(v) for v in MOLANG_PREFIXES):
                    self._controller_state[self._state_name]["on_exit"].append(f"{command};")
                else:
                    raise RuntimeError(
                        f"Invalid command for on_exit: {command}. Only Molang commands are allowed. Resource Pack Controller State [{self._state_name}]"
                    )

        self._default = False
        return self

    def animation(self, animation: str, condition: str = None):
        """Animation short name to play during this state.

        Parameters
        ----------
        animation : str
            The name of the animation to play.
        condition : str , optional
            The condition on which this animation plays.

        Returns
        -------
            This state.

        """
        if condition is None:
            self._controller_state[self._state_name]["animations"].append(animation)
        else:
            self._controller_state[self._state_name]["animations"].append({animation: condition})
        self._default = False
        return self

    def transition(self, state: str, condition: str):
        """Target state to switch to and the condition to do so.

        Parameters
        ----------
        state : str
            The name of the state to transition to.
        condition : str
            The condition that must be met for the transition to occur.

        Returns
        -------
            This state.

        """
        self._controller_state[self._state_name]["transitions"].append({state: str(condition)})
        self._default = False
        return self

    def particle(
        self,
        effect: str,
        locator: str = "root",
        pre_anim_script: str = None,
        bind_to_actor: bool = True,
    ):
        """The effect to be emitted during this state.

        Parameters
        ----------
        effect : str
            The shortname of the particle effect to be played, defined in the Client Entity.
        locator : str
            The name of a locator on the actor where the effect should be located.
        pre_anim_script : str , optional
            A molang script that will be run when the particle emitter is initialized.
        bind_to_actor : bool , optional
            Set to false to have the effect spawned in the world without being bound to an actor.

        Returns
        -------
            This state.

        """
        if not self._actor._description["description"]["particle_effects"].get(effect):
            raise RuntimeError(f"Particle effect '{effect}' is not defined in the Client Entity for Actor '{self._actor.name}'.")
        particle = {"effect": effect, "locator": locator}
        if pre_anim_script is not None:
            particle.update({"pre_effect_script": f"{pre_anim_script};".replace(";;", ";")})
        if bind_to_actor is False:
            particle.update({"bind_to_actor": False})
        self._controller_state[self._state_name]["particle_effects"].append(particle)
        self._default = False
        return self

    def sound_effect(
        self,
        effect: str,
        locator: str = "root",
    ):
        """Collection of sounds to trigger on entry to this animation state.

        Parameters
        ----------
        effect : str
            The shortname of the sound effect to be played, defined in the Client Entity.

        Returns
        -------
            This state.

        """
        self._controller_state[self._state_name]["sound_effects"].append({"effect": effect, "locator": locator})
        self._default = False
        return self

    def blend_transition(self, blend_value: float):
        """Sets the amount of time to fade out if the animation is interrupted.

        Parameters
        ----------
        blend_value : float
            Blend out time.

        Returns
        -------
            This state.

        """
        self._controller_state[self._state_name]["blend_transition"] = blend_value
        self._default = False
        return self

    @property
    def blend_via_shortest_path(self):
        """When blending a transition to another state, animate each euler axis through the shortest rotation, instead of by value.

        Returns
        -------
            This state.

        """
        self._controller_state[self._state_name]["blend_via_shortest_path"] = True
        self._default = False
        return self

    @property
    def _export(self):
        return self._controller_state


class _BP_Controller:
    def __init__(self, identifier, controller_shortname):
        self._identifier = identifier
        self._controller_shortname = controller_shortname
        self._controllers = JsonSchemes.animation_controller(self._identifier, self._controller_shortname)
        self._controller_states: list[_BP_ControllerState] = []
        self._controller_namespace = f"controller.animation.{self._identifier.replace(':', '.')}.{self._controller_shortname}"
        self._states_names = []
        self._side = "Server"

    def add_state(self, state_name: str):
        self._states_names.append(state_name)
        """Adds a new state to the animation controller.
        
        Parameters
        ----------
        state_name : str
            The name of the state to add.
        
        Returns
        -------
            Animation controller state.
        
        """
        self._controller_state = _BP_ControllerState(state_name)
        self._controller_states.append(self._controller_state)
        return self._controller_state

    @property
    def _export(self):
        collected_states = []
        for state in self._controller_states:
            if not state._default:
                self._controllers[self._controller_namespace]["states"].update(state._export)
                for tr in state._export.values():
                    if "transitions" in tr:
                        for st in tr["transitions"]:
                            collected_states.extend(st.keys())

        for state in set(collected_states):
            if state not in self._states_names:
                raise RuntimeError(
                    f"State '{state}' is referenced in transitions but not defined in the. behavior Pack Animation Controller[{self._identifier}]."
                )

        if len(self._controllers[self._controller_namespace]["states"].items()) > 0:
            return self._controllers
        return {}


class _RP_Controller(_BP_Controller):
    def __init__(self, actor: "_ActorClientDescription", name, controller_shortname):
        super().__init__(name, controller_shortname)
        self._actor = actor
        self._side = "Client"
        self._controller_states: list[_RP_ControllerState] = []

    def add_state(self, state_name: str):
        self._states_names.append(state_name)
        self._controller_state = _RP_ControllerState(self._actor, state_name)
        self._controller_states.append(self._controller_state)
        return self._controller_state


class _BP_AnimationControllers(AddonObject):
    _extension = ".bp_ac.json"
    _path = os.path.join(
        CONFIG.BP_PATH,
        "animation_controllers",
    )
    _object_type = "Behavior Pack Animation Controller"

    def __init__(self, identifier) -> None:
        super().__init__(identifier)
        self._animation_controllers = JsonSchemes.animation_controllers()
        self._controllers_list: list[_BP_Controller] = []

    def add_controller(self, controller_shortname: str) -> _BP_Controller:
        """Adds a new animation controller to the current actor with `default` as the `initial_state`.

        Parameters
        ----------
        controller_shortname : str
            The shortname of the controller you want to add.

        Returns
        -------
            Animation controller.

        """
        ctrl = _BP_Controller(self.identifier, controller_shortname)
        self._controllers_list.append(ctrl)
        return ctrl

    def queue(self, directory: str = None):
        if len(self._controllers_list) > 0:
            for controller in self._controllers_list:
                self._animation_controllers["animation_controllers"].update(controller._export)
            self.content(self._animation_controllers)
            return super().queue(directory=directory)


class _RP_AnimationControllers(AddonObject):
    _extension = ".rp_ac.json"
    _path = os.path.join(CONFIG.RP_PATH, "animation_controllers")
    _object_type = "Resource Pack Animation Controller"

    def __init__(self, actor: "_ActorClientDescription", name) -> None:
        super().__init__(name)
        self._actor = actor
        self._animation_controllers = JsonSchemes.animation_controllers()
        self._controllers_list: list[_RP_Controller] = []

    def add_controller(self, controller_shortname: str) -> _RP_Controller:
        """Adds a new animation controller to the current actor with `default` as the `initial_state`.

        Parameters
        ----------
        controller_shortname : str
            The shortname of the controller you want to add.

        Returns
        -------
            Animation controller.

        """
        self._animation_controller = _RP_Controller(self._actor, self.identifier, controller_shortname)
        self._controllers_list.append(self._animation_controller)
        return self._animation_controller

    def _is_populated(self):
        return len(self._controllers_list) > 0 and any([len(t._controller_states) > 0 for t in [c for c in self._controllers_list]])

    def queue(self, directory: str = None):
        if self._is_populated():
            for controller in self._controllers_list:
                self._animation_controllers["animation_controllers"].update(controller._export)
            self.content(self._animation_controllers)
            return super().queue(directory=directory)


class _ActorReuseAssets:
    def __init__(self, client: "_ActorClientDescription") -> None:
        self.client = client

    def animation(self, shortname: str, animation_name: str, animate: bool = False, condition: str | Molang = None):
        if animate is True:
            if condition is None:
                self.client._animate_append(shortname)
            else:
                self.client._animate_append({shortname: condition})

        self.client._description["description"]["animations"].update({shortname: animation_name})

    def animation_controller(self, shortname: str, animation_controller_name: str, animate: bool = False, condition: str | Molang = None):
        if animate is True:
            if condition is None:
                self.client._animate_append(shortname)
            else:
                self.client._animate_append({shortname: condition})

        self.client._description["description"]["animations"].update({shortname: animation_controller_name})

    def texture(self, shortname: str, texture_name: str):
        self.client._description["description"]["textures"].update({shortname: texture_name})

    def geometry(self, shortname: str, geometry_name: str):
        self.client._description["description"]["geometry"].update({shortname: geometry_name})

    def particle_effect(self, shortname: str, particle_name: str):
        self.client["description"]["particle_effects"].update({shortname: particle_name})

    def sound_effect(self, shortname: str, sound_name: str):
        self.client._description["description"]["sound_effects"].update({shortname: sound_name})

    def spawn_egg(self, item_sprite: str, texture_index: int = 0):
        self.client._description["description"]["spawn_egg"] = {"texture": item_sprite, "texture_index": texture_index}

    def render_controller(self, controller_name: str, condition: str = None):
        if condition is None:
            self.client._description["description"]["render_controllers"].append(controller_name)
        else:
            self.client._description["description"]["render_controllers"].append({controller_name: condition})


class _ActorDescription(MinecraftDescription):
    """Base class for all actor descriptions."""

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all actor descriptions.

        Parameters:
            name (str): The name of the actor.
            is_vanilla (bool, optional): Whether or not the actor is a vanilla actor. Defaults to False.
        """
        super().__init__(name, is_vanilla)
        self._animation_controllers: _RP_AnimationControllers
        self._description["description"].update({"animations": {}, "scripts": {"animate": []}})

    def _animate_append(self, key: str | dict):
        """Appends a key to the animate list."""

        if key not in self._description["description"]["scripts"]["animate"]:
            self._description["description"]["scripts"]["animate"].append(key)

    def _animation_controller(self, controller_shortname: str, animate: bool = False, condition: str = None):
        """Sets the mapping of internal animation controller references to actual animations.

        Parameters
        ----------
        controller_shortname : str
            The name of the animation controller.
        animate : bool, optional
            bool = False
        condition : str, optional
            str = None

        """

        if animate is True:
            if condition is None:
                self._animate_append(controller_shortname)
            else:
                self._animate_append({controller_shortname: condition})

        self._description["description"]["animations"].update(
            {controller_shortname: f"controller.animation.{CONFIG.NAMESPACE}.{self.name}.{controller_shortname}"}
        )

    def _animations(self, geometry_name: str, animation_shortname: str, animate: bool = False, condition: str = None):
        """Sets the mapping of internal animation references to actual animations.

        Parameters
        ----------
        animation_shortname : str
            The name of the animation.
        animate : bool, optional
            bool = False
        condition : str, optional
            str = None

        """

        if animate is True:
            if condition is None:
                self._animate_append(animation_shortname)
            else:
                self._animate_append({animation_shortname: condition})

        self._description["description"]["animations"].update(
            {animation_shortname: f"animation.{CONFIG.NAMESPACE}.{geometry_name}.{animation_shortname}"}
        )


class _ActorClientDescription(_ActorDescription):
    _queued_animation_controllers = set()
    _type = "entity"

    def _render_append(self, key):
        if key not in self._description["description"]["render_controllers"]:
            self._description["description"]["render_controllers"].append(key)

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all client actor descriptions.

        Parameters:
            name (str): The name of the actor.
            is_vanilla (bool, optional): Whether or not the actor is a vanilla actor. Defaults to False.
            type (str, optional): The type of the actor. Defaults to "entity".
        """
        super().__init__(name, is_vanilla)
        if self._type not in ["entity", "attachables"]:
            raise RuntimeError(
                f"Invalid type '{self._type}' for actor description. Expected 'entity' or 'attachables'. Actor [{self.identifier}]"
            )

        if is_vanilla and self.identifier not in list(map(str, MinecraftEntityTypes.__dict__.values())):
            raise RuntimeError(
                f"Invalid vanilla entity '{self.identifier}'. Please use a valid vanilla entity from MinecraftEntityTypes. Actor [{self.identifier}]"
            )

        self._is_dummy = False
        self._animation_controllers = _RP_AnimationControllers(self, self._name)
        self._render_controllers = _RenderControllers(self._name)
        self._sounds: list[SoundDescription] = []
        self._texture_set: __TextureSet = None

        self._description["description"].update(JsonSchemes.client_description())

    def animation_controller(self, controller_shortname: str, animate: bool = False, condition: str | Molang = None):
        """Sets the mapping of internal animation controller references to actual animations.

        Parameters:
            controller_shortname (str): The name of the animation controller.
            animate (bool, optional): Whether or not to animate the animation controller. Defaults to False.
            condition (str| Molang, optional): The condition to animate the animation controller. Defaults to None.

        """
        self._animation_controller(controller_shortname, animate, condition)
        return self._animation_controllers.add_controller(controller_shortname)

    def animation(self, blockbench_name: str, animation_name: str, animate: bool = False, condition: str | Molang = None):
        """Sets the mapping of internal animation references to actual animations.

        Parameters:
            animation_name (str): The name of the animation.
            animate (bool, optional): Whether or not to animate the animation. Defaults to False.
            condition (str | Molang, optional): The condition to animate the animation. Defaults to None.

        """
        bb = _Blockbench(blockbench_name, "actors")
        bb.animations.queue_animation(animation_name)

        self._animations(blockbench_name, animation_name, animate, condition)

        return self

    def material(self, material_id: str, material_name: str):
        """This method manages the materials for an entity.

        Parameters:
            material_id (str): The id of the material.
            material_name (str): The name of the material.

        """
        self._description["description"]["materials"].update({str(material_id): str(material_name)})
        return self

    def geometry(self, geometry_name: str, override_bounding_box: Vector2D = None):
        """
        This method manages the geometry for an entity.

        Parameters:
            geometry_name (str): The name of the geometry.
            override_bounding_box (Vector2D, optional): The bounding box to override the default bounding box. Defaults to None.

        Returns:
            self: Returns an instance of the class.
        """

        bb = _Blockbench(geometry_name, "actors")
        if override_bounding_box:
            bb.override_bounding_box(override_bounding_box)
        bb.model.queue_model()

        self._description["description"]["geometry"].update({geometry_name: f"geometry.{CONFIG.NAMESPACE}.{geometry_name}"})

        return self

    def dummy(self):
        """This method manages the dummy for an entity."""
        self._is_dummy = True
        return self

    def queryable_geometry(self, geometry_shortname: str):
        """This method manages the queryable geometry for an entity.

        Parameters:
            geometry_shortname (str): The shortname of the geometry.

        """
        if not geometry_shortname in self._description["description"]["geometry"]:
            raise RuntimeError(
                f"Queryable geometry '{geometry_shortname}' not found in entity {self.identifier}. Entity [{self.identifier}]"
            )
        self._description["description"]["queryable_geometry"] = geometry_shortname
        return self

    def texture(
        self,
        blockbench_name: str,
        color_texture: str,
        normal_texture: str | RGB | RGBA = None,
        heightmap_texture: str | RGB | RGBA = None,
        metalness_emissive_roughness_texture: str | RGB | RGBA = None,
        metalness_emissive_roughness_subsurface_texture: str | RGB | RGBA = None,
    ):
        """This method manages the textures for an entity.

        Parameters:
            blockbench_name (str): The name of the texture.
            texture_name (str): The name of the texture.
        """
        bb = _Blockbench(blockbench_name, "actors")
        bb.textures.queue_texture(color_texture)

        self._description["description"]["textures"].update(
            {color_texture: os.path.join("textures", CONFIG.NAMESPACE, CONFIG.PROJECT_NAME, "actors", blockbench_name, color_texture)}
        )

        if any(
            [
                normal_texture,
                heightmap_texture,
                metalness_emissive_roughness_texture,
                metalness_emissive_roughness_subsurface_texture,
            ]
        ):
            self._texture_set = __TextureSet(self.identifier, "entities")
            self._texture_set.set_textures(
                blockbench_name,
                color_texture,
                normal_texture,
                heightmap_texture,
                metalness_emissive_roughness_texture,
                metalness_emissive_roughness_subsurface_texture,
            )
        return self

    def script(self, variable: Variable | str, script: Molang | str):
        """This method manages the scripts for an entity."""
        self._description["description"]["scripts"]["pre_animation"].append(f"{variable}={script};".replace(";;", ";"))
        return self

    def parent_setup(self, variable: Variable | str, script: Molang | str):
        """This method manages the scripts for an entity."""
        self._description["description"]["scripts"]["parent_setup"].append(f"{variable}={script};".replace(";;", ";"))
        return self

    def init_vars(self, **vars):
        """Initializes variables for an entity.

        Examples:
            >>> Entity("example").init_vars(x=0, y=5, z=8)
        """
        for k, v in vars.items():
            Variable._set_var(k)
            self._description["description"]["scripts"]["initialize"].append(f"v.{k}={v};")
        return self

    def scale(self, scale: Molang | str = "1"):
        """Sets the scale of the entity.

        Parameters:
            scale (str | Molang, optional): The scale of the entity. Defaults to "1".
        """
        if not scale == "1":
            self._description["description"]["scripts"].update({"scale": str(scale)})

    def should_update_bones_and_effects_offscreen(self, bool: bool = False):
        """Sets whether or not the entity should update bones and effects offscreen.

        Parameters:
            bool (bool, optional): Whether or not the entity should update bones and effects offscreen. Defaults to False.
        """
        self._description["description"]["scripts"]["should_update_bones_and_effects_offscreen"] = str(int(bool))

    def should_update_effects_offscreen(self, bool: bool = False):
        """Sets whether or not the entity should update effects offscreen.

        Parameters:
            bool (bool, optional): Whether or not the entity should update effects offscreen. Defaults to False.
        """
        self._description["description"]["scripts"]["should_update_effects_offscreen"] = str(int(bool))

    def scaleXYZ(self, x: Molang | str = "1", y: Molang | str = "1", z: Molang | str = "1"):
        """Sets the scale of the entity.

        Parameters:
            x (str | Molang, optional): The x scale of the entity. Defaults to "1".
            y (str | Molang, optional): The y scale of the entity. Defaults to "1".
            z (str | Molang, optional): The z scale of the entity. Defaults to "1".
        """
        if not x == "1" or not y == "1" or not z == "1":
            self._description["description"]["scripts"]["scalex"] = str(x)
            self._description["description"]["scripts"]["scaley"] = str(y)
            self._description["description"]["scripts"]["scalez"] = str(z)

    def render_controller(self, controller_name: str, condition: str = None):
        """This method manages the render controllers for an entity.

        Parameters:
            controller_name (str): The name of the render controller.
            condition (str, optional): The condition to render the render controller. Defaults to None.

        """
        if condition is None:
            self._render_append(f"controller.render.{CONFIG.NAMESPACE}.{self.name}.{controller_name}")
        else:
            self._render_append({f"controller.render.{CONFIG.NAMESPACE}.{self.name}.{controller_name}": condition})

        return self._render_controllers.add_controller(controller_name)

    def particle_effect(self, particle_name: str, use_vanilla_texture: bool = False):
        """This method manages the particle effects for an entity.

        Parameters:
            particle_name (str): The name of the particle effect.

        """
        from anvil.api.world.particles import Particle

        self._particle_name = particle_name
        self._description["description"]["particle_effects"].update({self._particle_name: f"{CONFIG.NAMESPACE}:{self._particle_name}"})
        Particle(self._particle_name, use_vanilla_texture).queue()
        return self

    def sound_effect(
        self,
        sound_name: str,
        sound_path: str,
        category: SoundCategory = SoundCategory.Neutral,
        max_distance: int = 0,
        min_distance: int = 9999,
    ):
        """This method manages the sound effects for an entity.

        Parameters:
            sound_shortname (str): The shortname of the sound effect.
            sound_path (str): The path to the sound effects from the assets/sounds folder.
            category (SoundCategory, optional): The category of the sound effect. Defaults to SoundCategory.Neutral.

        """
        self._description["description"]["sound_effects"].update({sound_name: f"{CONFIG.NAMESPACE}:{sound_path}"})
        sound: SoundDescription = ANVIL.definitions.register_sound_definition(
            sound_path, category, max_distance=max_distance, min_distance=min_distance
        )
        self._sounds.append(sound)
        return sound

    def sound_event(
        self,
        sound_reference: str,
        sound_event: EntitySoundEvent,
        category: SoundCategory = SoundCategory.Neutral,
        volume: float = 1.0,
        pitch: tuple[float, float] = (0.8, 1.2),
        max_distance: int = 0,
        min_distance: int = 9999,
        variant_query: Molang = None,
        variant_map: str = None,
    ):
        """This method manages the sound events for an entity.

        Parameters:
            sound_reference (str): The identifier of the sound effect.
            sound_event (EntitySoundEvent): The sound event.
            category (SoundCategory, optional): The category of the sound effect. Defaults to SoundCategory.Neutral.
            volume (float, optional): The volume of the sound effect. Defaults to 1.0.
            pitch (tuple[float, float], optional): The pitch of the sound effect. Defaults to (0.8, 1.2).
            max_distance (int, optional): The maximum distance of the sound effect. Defaults to 0.
            min_distance (int, optional): The minimum distance of the sound effect. Defaults to 9999.
            variant_query (Molang, optional): The variant query of the sound effect. Defaults to None.
            variant_map (str, optional): The variant map of the sound effect. Defaults to None.
        """
        ANVIL.definitions.register_entity_sound_event(
            self.identifier,
            sound_reference,
            sound_event,
            category,
            volume,
            pitch,
            variant_query,
            variant_map,
            max_distance,
            min_distance,
        )

    def _export(self, directory: str = None):
        """Queues the entity for export.

        Parameters:
            directory (str, optional): The directory to export the entity to. Defaults to None.

        Raises:
            RuntimeError: If the entity does not have a geometry, texture, or render controller.
            Exception: If a geometry is reused but the entity it is reused from has not been queued yet.

        """
        if not self._is_dummy:
            if len(self._description["description"]["geometry"]) == 0:
                raise RuntimeError(f"Entity {self.identifier} missing at least one geometry. Entity [{self.identifier}]")

            if len(self._description["description"]["textures"]) == 0:
                raise RuntimeError(f"Entity {self.identifier} missing at least one texture. Entity [{self.identifier}]")

            if len(self._description["description"]["render_controllers"]) == 0:
                raise RuntimeError(f"Entity {self.identifier} missing at least one render controller. Entity [{self.identifier}]")

            for controller in self._render_controllers._controllers:
                controller._validate(
                    self._description["description"]["textures"].keys(),
                    self._description["description"]["geometry"].keys(),
                    self._description["description"]["materials"].keys(),
                )

            self._render_controllers.queue(directory)
            self._animation_controllers.queue(directory)

        for sound in self._sounds:
            sound._export

        if self._texture_set:
            self._texture_set.queue()

        if self._description["description"]["scripts"]["parent_setup"] != []:
            self._description["description"]["scripts"]["parent_setup"] = ";".join(
                self._description["description"]["scripts"]["parent_setup"]
            )

        for controller in self._animation_controllers._controllers_list:
            if controller._export == {}:
                self._description["description"]["animations"].pop(controller._controller_shortname)
                anims: list[str | dict] = self._description["description"]["scripts"]["animate"]

                if controller._controller_shortname in anims:
                    anims.remove(controller._controller_shortname)
                else:
                    d = next(d for d in anims if isinstance(d, dict) and controller._controller_shortname in d)
                    anims.remove(d)

        return super()._export()


class _EntityClientDescription(_ActorClientDescription):
    """Base class for all client entity descriptions."""

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all client entity descriptions.

        Parameters:
            name (str): The name of the entity.
            is_vanilla (bool, optional): Whether or not the entity is a vanilla entity. Defaults to False.
        """
        self._spawn_egg_texture = None
        super().__init__(name, is_vanilla)

    @property
    def EnableAttachables(self):
        """This determines if the entity should render attachables such as armor."""
        self._description["description"]["enable_attachables"] = True
        return self

    @property
    def HeldItemIgnoresLighting(self):
        """This determines if the held item should ignore lighting."""
        self._description["description"]["held_item_ignores_lighting"] = True
        return self

    @property
    def HideArmor(self):
        """This determines if the armor should be hidden."""
        self._description["description"]["hide_armor"] = True
        return self

    def spawn_egg(self, item_sprite: str, texture_index: int = 0):
        """This method adds a spawn egg texture to the entity.

        Parameters:
            item_sprite (str): The name of the item sprite.
        """
        ANVIL.definitions.register_item_textures(item_sprite, "spawn_eggs", item_sprite)
        self._description["description"]["spawn_egg"] = {
            "texture": f"{CONFIG.NAMESPACE}:{item_sprite}",
            "texture_index": texture_index if texture_index == 0 else {},
        }

    def spawn_egg_color(self, base_color: str, overlay_color: str):
        """This method adds a spawn egg color to the entity.

        Parameters:
            base_color (str): The base color of the spawn egg.
            overlay_color (str): The overlay color of the spawn egg.
        """
        self._description["description"]["spawn_egg"] = {"base_color": base_color, "overlay_color": overlay_color}

    def _export(self, directory: str):
        """Queues the entity for export.

        Parameters:
            directory (str): The directory to export the entity to.

        """
        super()._export(directory)
        if "spawn_egg" not in self._description["description"] and not self._is_vanilla:
            self._description["description"]["spawn_egg"] = {
                "base_color": "#FFFFFF",
                "overlay_color": "#000000",
            }

        return self._description


class _AttachableClientDescription(_ActorClientDescription):
    """Base class for all client attachable descriptions."""

    _type = "attachables"

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all client attachable descriptions.

        Parameters:
            name (str): The name of the attachable.
            is_vanilla (bool, optional): Whether or not the attachable is a vanilla attachable. Defaults to False.
        """
        super().__init__(name, is_vanilla)

    @property
    def reuse_assets(self):
        """Whether or not the actor should reuse assets from another actor."""
        return _ActorReuseAssets(self._description)


class _EntityServerDescription(_ActorDescription):
    """Base class for all server entity descriptions."""

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all server entity descriptions.

        Parameters:
            name (str): The name of the entity.
            is_vanilla (bool, optional): Whether or not the entity is a vanilla entity. Defaults to False.
        """
        super().__init__(name, is_vanilla)
        self._properties = _Properties()
        self._description["description"].update(
            {
                "properties": {},
            }
        )

    @property
    def Summonable(self):
        """Sets whether or not we can summon this entity using commands such as /summon.

        Returns
        -------
            Entity description object.

        """
        self._description["description"]["is_summonable"] = True
        return self

    @property
    def Spawnable(self):
        """Sets whether or not this entity has a spawn egg in the creative ui.

        Returns
        -------
            Entity description object.

        """
        self._description["description"]["is_spawnable"] = True
        return self

    @property
    def Experimental(self):
        """Sets whether or not this entity is experimental. Experimental entities are only enabled when the experimental toggle is enabled.

        Returns
        -------
            Entity description object.

        """
        if CONFIG._EXPERIMENTAL:
            if CONFIG._TARGET == "addon":
                raise RuntimeError(f"Experimental entities are not allowed for packages of type 'addon'. Entity [{self.identifier}].")

        self._description["description"]["is_experimental"] = True
        return self

    def RuntimeIdentifier(self, entity: "MinecraftEntityDescriptor"):
        """Sets the runtime identifier of the entity.

        Parameters:
            entity (MinecraftEntityDescriptor): The vanilla entity to get the runtime identifier from.
        """
        if CONFIG._TARGET != ConfigPackageTarget.ADDON:
            if type(entity) is MinecraftEntityDescriptor:
                if entity._allow_runtime:
                    self._description["description"]["runtime_identifier"] = entity.identifier
                else:
                    raise RuntimeError(f"Entity {entity.identifier} does not allow runtime identifier usage. Entity [{self.identifier}].")
            else:
                raise TypeError(f"Expected Entity, got {type(entity).__name__}. Entity [{self.identifier}]")
        else:
            raise RuntimeError(f"Using runtime is not allowed for packages of type '{CONFIG._TARGET}'. Entity [{self.identifier}]")

    @property
    def add_property(self):
        """Adds a property to the entity."""
        return self._properties

    def spawn_category(self):
        """Sets the spawn category of the entity."""
        self._description["description"]["spawn_category"] = "none"

    @property
    def _export(self):
        """Exports the entity description."""
        self._description["description"]["properties"] = self._properties._export
        return super()._export()


class _EntityServer(AddonObject):
    """Base class for all server entities."""

    _extension = ".behavior.json"
    _path = os.path.join(CONFIG.BP_PATH, "entities")
    _object_type = "Server Entity"

    def _add_despawn_function(self):
        """Adds a despawn function to the entity."""
        self.component_group("despawn").add(EntityInstantDespawn())
        self.event("despawn").add("despawn")

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all server entities.

        Parameters:
            name (str): The name of the entity.
            is_vanilla (bool, optional): Whether or not the entity is a vanilla entity. Defaults to False.
        """
        super().__init__(name, is_vanilla)
        self._server_entity = JsonSchemes.entity_server()
        self._description = _EntityServerDescription(self.identifier, self._is_vanilla)
        self._animation_controllers = _BP_AnimationControllers(self.identifier)
        self._animations = _BPAnimations(self.identifier)
        self._spawn_rule = SpawnRule(self.identifier, self._is_vanilla)
        self._components = _Components()
        self._events: list[_Event] = []
        self._component_groups: list[_ComponentGroup] = []
        self._vars = []
        self._add_despawn_function()

    @property
    def description(self):
        """Returns the entity description."""
        return self._description

    @property
    def spawn_rule(self):
        """Returns the spawn rule of the entity."""
        return self._spawn_rule

    def animation_controller(self, controller_shortname: str, animate: bool = False, condition: str | Molang = None):
        """Sets the mapping of internal animation controller references to actual animations.

        Parameters:
            controller_shortname (str): The name of the animation controller.
            animate (bool, optional): Whether or not to animate the animation controller. Defaults to False.
            condition (str | Molang, optional): The condition to animate the animation controller. Defaults to None.

        """
        self._description._animation_controller(controller_shortname, animate, condition)
        return self._animation_controllers.add_controller(controller_shortname)

    def animation(
        self,
        animation_name: str,
        loop: bool = False,
        animate: bool = False,
        condition: str | Molang = None,
    ):
        """Sets the mapping of internal animation references to actual animations.

        Parameters:
            animation_name (str): The name of the animation.
            loop (bool, optional): Whether or not the animation should loop. Defaults to False.
            animate (bool, optional): Whether or not to animate the animation. Defaults to False.
            condition (str | Molang, optional): The condition to animate the animation. Defaults to None.

        """
        self._description._animations(self._name, animation_name, animate, condition)
        return self._animations.add_animation(animation_name, loop)

    def init_vars(self, **vars):
        """Initializes variables for an entity."""
        for k, v in vars.items():
            Variable._set_var(k)
            self._vars.extend([f"v.{k}={v}"])

        return self

    def event(self, event_name: str):
        """Adds an event to the entity.

        Parameters:
            event_name (str): The name of the event.
        """
        self._events.append(_Event(event_name))
        return self._events[-1]

    @property
    def components(self):
        """Returns the components of the entity."""
        return self._components

    def component_group(self, component_group_name: str):
        """Adds a component group to the entity.

        Parameters:
            component_group_name (str): The name of the component group.

        """
        self._component_groups.append(_ComponentGroup(component_group_name))
        return self._component_groups[-1]

    def add_basic_components(self):
        """Adds basic server components to the entity.

        This includes:
            - `EntityBreathable`
            - `EntityCollisionBox`
            - `EntityDamageSensor`
            - `EntityHealth`
            - `EntityJumpStatic`
            - `EntityKnockbackResistance`
            - `EntityMovement`
            - `EntityMovementType`
            - `EntityNavigationType`
            - `EntityPhysics`
            - `EntityPushable`
            - `EntityPushThrough`
        """
        from anvil.api.actors.components import (EntityBreathable,
                                                 EntityCollisionBox,
                                                 EntityDamageSensor,
                                                 EntityHealth,
                                                 EntityJumpStatic,
                                                 EntityKnockbackResistance,
                                                 EntityMovement,
                                                 EntityMovementType,
                                                 EntityNavigationType,
                                                 EntityPhysics, EntityPushable,
                                                 EntityPushThrough)
        from anvil.api.logic.commands import DamageCause

        self.components.add(
            EntityJumpStatic(0),
            EntityMovementType.Basic(),
            EntityNavigationType.Walk(),
            EntityMovement(0),
            EntityPhysics(True, True),
            EntityKnockbackResistance(10000),
            EntityHealth(6),
            EntityCollisionBox(1, 1),
            EntityBreathable(),
            EntityDamageSensor().add_trigger(DamageCause.All, DamageSensor.No),
            EntityPushable(False, False),
            EntityPushThrough(1),
        )

    def queue(self, directory: str = None):
        """Queues the entity for export.

        Parameters:
            directory (str, optional): The directory to export the entity to. Defaults to None.
        """
        if len(self._vars) > 0:
            self.animation_controller("variables", True).add_state("default").on_entry(*self._vars)
        self._animations.queue(directory=directory)
        self._animation_controllers.queue(directory=directory)
        self._spawn_rule.queue(directory=directory)

        self._server_entity["minecraft:entity"].update(self.description._export)
        self._server_entity["minecraft:entity"]["components"].update(self._components._export()["components"])

        for event in self._events:
            self._server_entity["minecraft:entity"]["events"].update(event._export)
        for component_group in self._component_groups:
            self._server_entity["minecraft:entity"]["component_groups"].update(component_group._export())

        self.content(self._server_entity)

        controllers = list(self._animation_controllers._animation_controllers["animation_controllers"].keys())
        cleared_items = []
        for key, controller in self._description._description["description"]["animations"].items():
            if controller.startswith("controller.") and controller not in controllers:
                cleared_items.append(key)

        for item in cleared_items:
            self._description._description["description"]["animations"].pop(item)

        if self._components._has(EntityRideable) or any(c._has(EntityRideable) for c in self._component_groups):
            ANVIL.definitions.register_lang(f"action.hint.exit.{self.identifier}", "Sneak to exit")

        super().queue(directory=directory)


class _EntityClient(AddonObject):
    """Base class for all client entities."""

    _extension = ".entity.json"
    _path = os.path.join(CONFIG.RP_PATH, "entity")
    _object_type = "Client Entity"

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Base class for all client entities.

        Parameters:
            name (str): The name of the entity.
            is_vanilla (bool, optional): Whether or not the entity is a vanilla entity. Defaults to False.
        """
        super().__init__(name, is_vanilla)
        self._client_entity = JsonSchemes.entity_client()
        self._description = _EntityClientDescription(self.identifier, self._is_vanilla)

    @property
    def reuse_assets(self):
        """Whether or not the actor should reuse assets from another actor."""
        return _ActorReuseAssets(self._description)

    @property
    def description(self):
        """Returns the entity description."""
        return self._description

    def queue(self, directory: str = None):
        """Queues the entity for export.

        Parameters:
            directory (str, optional): The directory to export the entity to. Defaults to None.
        """
        self._client_entity["minecraft:client_entity"].update(self._description._export(directory))

        self.content(self._client_entity)
        super().queue(directory=directory)


class Entity(EntityDescriptor):
    def __init__(self, name, is_vanilla=False, allow_runtime=True):
        super().__init__(name, is_vanilla, allow_runtime)

        self.server = _EntityServer(name, is_vanilla)
        self.client = _EntityClient(name, is_vanilla)

    def queue(
        self,
        directory: str = None,
        display_name: str = None,
        spawn_egg_name: str = None,
    ):
        display_name = self._display_name if display_name is None else display_name
        spawn_egg_name = f"Spawn {display_name}" if spawn_egg_name is None else spawn_egg_name
        if self._is_vanilla:
            directory = "vanilla"

        ANVIL.definitions.register_lang(f"entity.{self.identifier}.name", display_name)
        ANVIL.definitions.register_lang(f"entity.{self.identifier}<>.name", display_name)
        ANVIL.definitions.register_lang(
            f"item.spawn_egg.entity.{self.identifier}.name",
            spawn_egg_name,
        )
        self.client.queue(directory)
        self.server.queue(directory)

        CONFIG.Report.add_report(
            ReportType.ENTITY,
            vanilla=self._is_vanilla,
            col0=display_name,
            col1=self.identifier,
            col2=[event._event_name for event in self.server._events],
        )


class Attachable(AddonObject):
    _extension = ".attachable.json"
    _path = os.path.join(CONFIG.RP_PATH, "attachables")
    _object_type = "Attachable"

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """Initializes the attachable.

        Parameters:
            name (str): The name of the attachable.
        """
        super().__init__(name, is_vanilla)
        if not is_vanilla:
            self._is_vanilla = self.identifier in MinecraftItemTypes

        self._attachable = JsonSchemes.attachable()
        self._description = _AttachableClientDescription(self.identifier, False)

    @property
    def description(self):
        """Returns the description of the attachable."""
        return self._description

    def queue(self):
        """Queues the attachable."""

        self._attachable["minecraft:attachable"].update(self._description._export())
        self.content(self._attachable)

        CONFIG.Report.add_report(
            ReportType.ATTACHABLE,
            vanilla=self._is_vanilla,
            col0=self._display_name,
            col1=self.identifier,
        )
        super().queue()

    def __str__(self):
        """Returns the identifier of the attachable."""
        return self.identifier



from anvil.lib.schemas import MinecraftEntityDescriptor

class MinecraftEntityTypes:
    Agent = MinecraftEntityDescriptor("minecraft:agent", True, True)
    Allay = MinecraftEntityDescriptor("minecraft:allay", True, True)
    AreaEffectCloud = MinecraftEntityDescriptor("minecraft:area_effect_cloud", True, True)
    Armadillo = MinecraftEntityDescriptor("minecraft:armadillo", True, True)
    ArmorStand = MinecraftEntityDescriptor("minecraft:armor_stand", True, True)
    Arrow = MinecraftEntityDescriptor("minecraft:arrow", True, True)
    Axolotl = MinecraftEntityDescriptor("minecraft:axolotl", True, True)
    Bat = MinecraftEntityDescriptor("minecraft:bat", True, True)
    Bee = MinecraftEntityDescriptor("minecraft:bee", True, True)
    Blaze = MinecraftEntityDescriptor("minecraft:blaze", True, True)
    Boat = MinecraftEntityDescriptor("minecraft:boat", True, True)
    Bogged = MinecraftEntityDescriptor("minecraft:bogged", True, True)
    Breeze = MinecraftEntityDescriptor("minecraft:breeze", True, True)
    BreezeWindChargeProjectile = MinecraftEntityDescriptor("minecraft:breeze_wind_charge_projectile", True, True)
    Camel = MinecraftEntityDescriptor("minecraft:camel", True, True)
    Cat = MinecraftEntityDescriptor("minecraft:cat", True, True)
    CaveSpider = MinecraftEntityDescriptor("minecraft:cave_spider", True, True)
    ChestBoat = MinecraftEntityDescriptor("minecraft:chest_boat", True, True)
    ChestMinecart = MinecraftEntityDescriptor("minecraft:chest_minecart", True, True)
    Chicken = MinecraftEntityDescriptor("minecraft:chicken", True, True)
    Cod = MinecraftEntityDescriptor("minecraft:cod", True, True)
    CommandBlockMinecart = MinecraftEntityDescriptor("minecraft:command_block_minecart", True, True)
    Cow = MinecraftEntityDescriptor("minecraft:cow", True, True)
    Creaking = MinecraftEntityDescriptor("minecraft:creaking", True, True)
    Creeper = MinecraftEntityDescriptor("minecraft:creeper", True, True)
    Dolphin = MinecraftEntityDescriptor("minecraft:dolphin", True, True)
    Donkey = MinecraftEntityDescriptor("minecraft:donkey", True, True)
    DragonFireball = MinecraftEntityDescriptor("minecraft:dragon_fireball", True, True)
    Drowned = MinecraftEntityDescriptor("minecraft:drowned", True, True)
    Egg = MinecraftEntityDescriptor("minecraft:egg", True, True)
    ElderGuardian = MinecraftEntityDescriptor("minecraft:elder_guardian", True, True)
    EnderCrystal = MinecraftEntityDescriptor("minecraft:ender_crystal", True, True)
    EnderDragon = MinecraftEntityDescriptor("minecraft:ender_dragon", True, True)
    EnderPearl = MinecraftEntityDescriptor("minecraft:ender_pearl", True, True)
    Enderman = MinecraftEntityDescriptor("minecraft:enderman", True, True)
    Endermite = MinecraftEntityDescriptor("minecraft:endermite", True, True)
    EvocationIllager = MinecraftEntityDescriptor("minecraft:evocation_illager", True, True)
    EyeOfEnderSignal = MinecraftEntityDescriptor("minecraft:eye_of_ender_signal", True, True)
    Fireball = MinecraftEntityDescriptor("minecraft:fireball", True, True)
    FireworksRocket = MinecraftEntityDescriptor("minecraft:fireworks_rocket", True, True)
    FishingHook = MinecraftEntityDescriptor("minecraft:fishing_hook", True, True)
    Fox = MinecraftEntityDescriptor("minecraft:fox", True, True)
    Frog = MinecraftEntityDescriptor("minecraft:frog", True, True)
    Ghast = MinecraftEntityDescriptor("minecraft:ghast", True, True)
    GlowSquid = MinecraftEntityDescriptor("minecraft:glow_squid", True, True)
    Goat = MinecraftEntityDescriptor("minecraft:goat", True, True)
    Guardian = MinecraftEntityDescriptor("minecraft:guardian", True, True)
    HappyGhast = MinecraftEntityDescriptor("minecraft:happy_ghast", True, True)
    Hoglin = MinecraftEntityDescriptor("minecraft:hoglin", True, True)
    HopperMinecart = MinecraftEntityDescriptor("minecraft:hopper_minecart", True, True)
    Horse = MinecraftEntityDescriptor("minecraft:horse", True, True)
    Husk = MinecraftEntityDescriptor("minecraft:husk", True, True)
    IronGolem = MinecraftEntityDescriptor("minecraft:iron_golem", True, True)
    LightningBolt = MinecraftEntityDescriptor("minecraft:lightning_bolt", True, True)
    LingeringPotion = MinecraftEntityDescriptor("minecraft:lingering_potion", True, True)
    Llama = MinecraftEntityDescriptor("minecraft:llama", True, True)
    LlamaSpit = MinecraftEntityDescriptor("minecraft:llama_spit", True, True)
    MagmaCube = MinecraftEntityDescriptor("minecraft:magma_cube", True, True)
    Minecart = MinecraftEntityDescriptor("minecraft:minecart", True, True)
    Mooshroom = MinecraftEntityDescriptor("minecraft:mooshroom", True, True)
    Mule = MinecraftEntityDescriptor("minecraft:mule", True, True)
    Npc = MinecraftEntityDescriptor("minecraft:npc", True, True)
    Ocelot = MinecraftEntityDescriptor("minecraft:ocelot", True, True)
    OminousItemSpawner = MinecraftEntityDescriptor("minecraft:ominous_item_spawner", True, True)
    Panda = MinecraftEntityDescriptor("minecraft:panda", True, True)
    Parrot = MinecraftEntityDescriptor("minecraft:parrot", True, True)
    Phantom = MinecraftEntityDescriptor("minecraft:phantom", True, True)
    Pig = MinecraftEntityDescriptor("minecraft:pig", True, True)
    Piglin = MinecraftEntityDescriptor("minecraft:piglin", True, True)
    PiglinBrute = MinecraftEntityDescriptor("minecraft:piglin_brute", True, True)
    Pillager = MinecraftEntityDescriptor("minecraft:pillager", True, True)
    Player = MinecraftEntityDescriptor("minecraft:player", True, True)
    PolarBear = MinecraftEntityDescriptor("minecraft:polar_bear", True, True)
    Pufferfish = MinecraftEntityDescriptor("minecraft:pufferfish", True, True)
    Rabbit = MinecraftEntityDescriptor("minecraft:rabbit", True, True)
    Ravager = MinecraftEntityDescriptor("minecraft:ravager", True, True)
    Salmon = MinecraftEntityDescriptor("minecraft:salmon", True, True)
    Sheep = MinecraftEntityDescriptor("minecraft:sheep", True, True)
    Shulker = MinecraftEntityDescriptor("minecraft:shulker", True, True)
    ShulkerBullet = MinecraftEntityDescriptor("minecraft:shulker_bullet", True, True)
    Silverfish = MinecraftEntityDescriptor("minecraft:silverfish", True, True)
    Skeleton = MinecraftEntityDescriptor("minecraft:skeleton", True, True)
    SkeletonHorse = MinecraftEntityDescriptor("minecraft:skeleton_horse", True, True)
    Slime = MinecraftEntityDescriptor("minecraft:slime", True, True)
    SmallFireball = MinecraftEntityDescriptor("minecraft:small_fireball", True, True)
    Sniffer = MinecraftEntityDescriptor("minecraft:sniffer", True, True)
    SnowGolem = MinecraftEntityDescriptor("minecraft:snow_golem", True, True)
    Snowball = MinecraftEntityDescriptor("minecraft:snowball", True, True)
    Spider = MinecraftEntityDescriptor("minecraft:spider", True, True)
    SplashPotion = MinecraftEntityDescriptor("minecraft:splash_potion", True, True)
    Squid = MinecraftEntityDescriptor("minecraft:squid", True, True)
    Stray = MinecraftEntityDescriptor("minecraft:stray", True, True)
    Strider = MinecraftEntityDescriptor("minecraft:strider", True, True)
    Tadpole = MinecraftEntityDescriptor("minecraft:tadpole", True, True)
    ThrownTrident = MinecraftEntityDescriptor("minecraft:thrown_trident", True, True)
    Tnt = MinecraftEntityDescriptor("minecraft:tnt", True, True)
    TntMinecart = MinecraftEntityDescriptor("minecraft:tnt_minecart", True, True)
    TraderLlama = MinecraftEntityDescriptor("minecraft:trader_llama", True, True)
    TripodCamera = MinecraftEntityDescriptor("minecraft:tripod_camera", True, True)
    Tropicalfish = MinecraftEntityDescriptor("minecraft:tropicalfish", True, True)
    Turtle = MinecraftEntityDescriptor("minecraft:turtle", True, True)
    Vex = MinecraftEntityDescriptor("minecraft:vex", True, True)
    Villager = MinecraftEntityDescriptor("minecraft:villager", True, True)
    VillagerV2 = MinecraftEntityDescriptor("minecraft:villager_v2", True, True)
    Vindicator = MinecraftEntityDescriptor("minecraft:vindicator", True, True)
    WanderingTrader = MinecraftEntityDescriptor("minecraft:wandering_trader", True, True)
    Warden = MinecraftEntityDescriptor("minecraft:warden", True, True)
    WindChargeProjectile = MinecraftEntityDescriptor("minecraft:wind_charge_projectile", True, True)
    Witch = MinecraftEntityDescriptor("minecraft:witch", True, True)
    Wither = MinecraftEntityDescriptor("minecraft:wither", True, True)
    WitherSkeleton = MinecraftEntityDescriptor("minecraft:wither_skeleton", True, True)
    WitherSkull = MinecraftEntityDescriptor("minecraft:wither_skull", True, True)
    WitherSkullDangerous = MinecraftEntityDescriptor("minecraft:wither_skull_dangerous", True, True)
    Wolf = MinecraftEntityDescriptor("minecraft:wolf", True, True)
    XpBottle = MinecraftEntityDescriptor("minecraft:xp_bottle", True, True)
    XpOrb = MinecraftEntityDescriptor("minecraft:xp_orb", True, True)
    Zoglin = MinecraftEntityDescriptor("minecraft:zoglin", True, True)
    Zombie = MinecraftEntityDescriptor("minecraft:zombie", True, True)
    ZombieHorse = MinecraftEntityDescriptor("minecraft:zombie_horse", True, True)
    ZombiePigman = MinecraftEntityDescriptor("minecraft:zombie_pigman", True, True)
    ZombieVillager = MinecraftEntityDescriptor("minecraft:zombie_villager", True, True)
    ZombieVillagerV2 = MinecraftEntityDescriptor("minecraft:zombie_villager_v2", True, True)



from __future__ import annotations

import json
import os
import traceback
import uuid
from typing import Any, Dict, List, Mapping

from packaging.version import Version

from anvil import CONFIG
from anvil.lib.config import ConfigPackageTarget
from anvil.lib.enums import ComponentTarget
from anvil.lib.format_versions import *
from anvil.lib.lib import APPDATA, File, salt_from_str
from anvil.lib.templater import load_file
from anvil.lib.types import Identifier


class JsonSchemes:
    """A class used to read and write to the json_schemes.json file."""

    @staticmethod
    def pack_name_lang(name, description):
        return load_file("pack_name_lang.jsont", {"name": name, "description": description}).splitlines()

    @staticmethod
    def skin_pack_name_lang(name, display_name):
        return load_file("skin_pack_name_lang.jsont", {"name": name, "display_name": display_name}).splitlines()

    @staticmethod
    def esbuild_config_js(outDir, minify):
        return load_file("esbuild.jsont", {"out_dir": os.path.join(outDir, "scripts"), "minify": json.dumps(minify)}, is_json=False)

    @staticmethod
    def manifest_bp(version):
        m = load_file(
            "manifest_bp.jsont",
            {
                "bp_uuid": CONFIG._BP_UUID[0],
                "version": version,
                "engine_version": [int(i) for i in MANIFEST_BUILD.split(".")],
                "data_module_uuid": CONFIG._DATA_MODULE_UUID,
                "rp_uuid": CONFIG._RP_UUID[0],
                "company": CONFIG.COMPANY,
            },
            is_json=True,
        )

        if CONFIG._SCRIPT_API:
            m["modules"].append(
                {
                    "uuid": CONFIG._SCRIPT_MODULE_UUID,
                    "version": version,
                    "type": "script",
                    "language": "javascript",
                    "entry": "scripts/main.js",
                }
            )
            m["dependencies"].append(
                {
                    "module_name": "@minecraft/server",
                    "version": MODULE_MINECRAFT_SERVER,
                }
            )
            if CONFIG._SCRIPT_UI:
                m["dependencies"].append(
                    {
                        "module_name": "@minecraft/server-ui",
                        "version": MODULE_MINECRAFT_SERVER_UI,
                    }
                )
        if CONFIG._TARGET == "addon":
            m["header"]["pack_scope"] = "world"
            m["metadata"]["product_type"] = "addon"
        return m

    @staticmethod
    def manifest_rp(version):
        m = load_file(
            "manifest_rp.jsont",
            {
                "rp_uuid": CONFIG._RP_UUID[0],
                "version": version,
                "engine_version": [int(i) for i in MANIFEST_BUILD.split(".")],
                "resource_module_uuid": str(uuid.uuid4()),
                "bp_uuid": CONFIG._BP_UUID[0],
                "company": CONFIG.COMPANY,
            },
            is_json=True,
        )
        if CONFIG._PBR:
            m.update({"capabilities": ["pbr"]})
        if CONFIG._TARGET == "addon":
            m["header"]["pack_scope"] = "world"
            m["metadata"]["product_type"] = "addon"
        return m

    @staticmethod
    def manifest_world(version):
        m = load_file(
            "manifest_rp.jsont",
            {
                "world_uuid": CONFIG._PACK_UUID,
                "version": version,
                "engine_version": [int(i) for i in MANIFEST_BUILD.split(".")],
                "world_module_uuid": str(uuid.uuid4()),
                "company": CONFIG.COMPANY,
            },
            is_json=True,
        )

        if CONFIG._RANDOM_SEED:
            m["header"]["allow_random_seed"] = True

        return m

    @staticmethod
    def world_packs(version, pack_ids):
        return [{"pack_id": pack, "version": version} for pack in pack_ids]

    @staticmethod
    def skin_pack(serialize_name):
        return load_file("skin_pack.jsont", {"serialize_name": serialize_name}, is_json=True)

    @staticmethod
    def skin_json(filename: str, is_slim: bool, free: bool):
        return load_file(
            "skin_json.jsont",
            {
                "filename": filename,
                "geo_type": "customSlim" if is_slim else "custom",
                "paid": "free" if free else "paid",
            },
            is_json=True,
        )

    @staticmethod
    def manifest_skins(version):
        return load_file(
            "manifest_skins.jsont",
            {
                "skin_uuid": str(uuid.uuid4()),
                "version": version,
                "skin_module_uuid": str(uuid.uuid4()),
            },
            is_json=True,
        )

    @staticmethod
    def description(namespace, identifier):
        return {"description": {"identifier": f"{namespace}:{identifier}"}}

    @staticmethod
    def item_texture(resource_pack_name):
        return load_file("item_texture.jsont", {"resource_pack_name": resource_pack_name}, is_json=True)

    @staticmethod
    def sound_definitions():
        return load_file("sound_definitions.jsont", {"format_version": SOUND_DEFINITIONS_VERSION}, is_json=True)

    @staticmethod
    def music_definitions():
        return load_file("music_definitions.jsont", is_json=True)

    @staticmethod
    def sound(name, category):
        return {name: {"category": category, "sounds": []}}

    @staticmethod
    def materials():
        return load_file("materials.jsont", {"version": MATERIALS_VERSION}, is_json=True)

    @staticmethod
    def languages() -> Dict:
        return load_file("languages.jsont", is_json=True)

    @staticmethod
    def client_description():
        return {
            "materials": {"default": "entity_alphatest"},
            "scripts": {"pre_animation": [], "initialize": [], "animate": [], "parent_setup": []},
            "textures": {},
            "geometry": {},
            "particle_effects": {},
            "sound_effects": {},
            "render_controllers": [],
        }

    @staticmethod
    def entity_client():
        return load_file("entity_client.jsont", {"format_version": ENTITY_CLIENT_VERSION}, is_json=True)

    @staticmethod
    def entity_server():
        return load_file("entity_server.jsont", {"format_version": ENTITY_SERVER_VERSION}, is_json=True)

    @staticmethod
    def animations_bp():
        return load_file("animations_bp.jsont", {"format_version": BP_ANIMATION_VERSION}, is_json=True)

    @staticmethod
    def bp_animation(identifier, animation_short_name, loop):
        return {
            f"animation.{identifier.replace(':', '.')}.{animation_short_name}": {
                "loop": loop,
                "timeline": {},
            }
        }

    @staticmethod
    def animations_rp():
        return load_file("animations_rp.jsont", {"format_version": RP_ANIMATION_VERSION}, is_json=True)

    @staticmethod
    def animation_controller_state(state):
        return {
            state: {
                "on_entry": [],
                "on_exit": [],
                "animations": [],
                "transitions": [],
            }
        }

    @staticmethod
    def animation_controller(identifier, controller_shortname):
        return {
            f"controller.animation.{identifier.replace(':', '.')}.{controller_shortname}": {
                "initial_state": "default",
                "states": {},
            }
        }

    @staticmethod
    def animation_controllers():
        return load_file("animation_controllers.jsont", {"format_version": ANIMATION_CONTROLLERS_VERSION}, is_json=True)

    @staticmethod
    def geometry(model_name: str, texture_size: list[int], visible_box: list[int], visible_offset: list[int]):
        return load_file("geometry.jsont",
            {
                "format_version": GEOMETRY_VERSION,
                "namespace": CONFIG.NAMESPACE,
                "model_name": model_name,
                "texture_width": texture_size[0],
                "texture_height": texture_size[1],
                "visible_bounds_width": visible_box[0],
                "visible_bounds_height": visible_box[1],
                "visible_bounds_offset": visible_offset,
            },
            is_json=True,
        )

    @staticmethod
    def render_controller(identifier, controller_name):
        return {
            f"controller.render.{identifier.replace(':', '.')}.{controller_name}": {
                "arrays": {"textures": {}, "geometries": {}},
                "materials": [],
                "geometry": {},
                "textures": [],
                "part_visibility": [],
            }
        }

    @staticmethod
    def render_controllers():
        return load_file("render_controllers.jsont", {"format_version": RENDER_CONTROLLER_VERSION}, is_json=True)

    @staticmethod
    def attachable():
        return load_file("attachable.jsont", {"format_version": ENTITY_CLIENT_VERSION}, is_json=True)

    @staticmethod
    def spawn_rules():
        return load_file("spawn_rules.jsont", {"format_version": SPAWN_RULES_VERSION}, is_json=True)

    @staticmethod
    def server_block():
        return load_file("server_block.jsont", {"format_version": BLOCK_SERVER_VERSION}, is_json=True)

    @staticmethod
    def terrain_texture(resource_pack_name):
        return load_file("terrain_texture.jsont", {"resource_pack_name": resource_pack_name}, is_json=True)

    @staticmethod
    def flipbook_textures():
        return load_file("flipbook_textures.jsont", is_json=True)

    @staticmethod
    def font(font_name, font_file):
        return load_file("font.jsont", {"font_name": font_name, "font_file": font_file}, is_json=True)

    @staticmethod
    def fog():
        return load_file("fog.jsont", {"format_version": FOG_VERSION}, is_json=True)

    @staticmethod
    def dialogues():
        return load_file("dialogues_json.jsont", {"format_version": DIALOGUE_VERSION}, is_json=True)

    @staticmethod
    def dialogue_scene(scene_tag, npc_name, text, on_open_commands, on_close_commands, buttons):
        return {
            "scene_tag": scene_tag,
            "npc_name": npc_name,
            "text": text,
            "on_open_commands": on_open_commands,
            "on_close_commands": on_close_commands,
            "buttons": buttons,
        }

    @staticmethod
    def dialogue_button(name, commands):
        return {"name": name, "commands": commands}

    @staticmethod
    def server_item():
        return load_file("server_item.jsont", {"format_version": ITEM_SERVER_VERSION}, is_json=True)

    @staticmethod
    def camera_preset(identifier, inherit_from):
        return load_file(
            "camera_preset.jsont", {"format_version": CAMERA_PRESET_VERSION, "identifier": identifier, "inherit_from": inherit_from}, is_json=True
        )

    @staticmethod
    def package_json(project_name, version, description, author):
        return load_file(
            "package_json.jsont", {"project_name": project_name, "version": version, "description": description, "author": author}, is_json=True
        )

    @staticmethod
    def tsconstants(namespace: str, project_name: str):
        return load_file("tsconstants.jsont", {"namespace": namespace, "project_name": project_name})

    @staticmethod
    def tsconfig(out_dir):
        return load_file("tsconfig.jsont", {"out_dir": os.path.join(out_dir, "scripts")}, is_json=True)

    @staticmethod
    def blocks_json():
        return load_file("blocks_json.jsont", {"format_version": [int(i) for i in BLOCK_JSON_FORMAT_VERSION.split(".")]}, is_json=True)

    @staticmethod
    def sounds_json():
        return load_file("sounds_json.jsont", is_json=True)

    @staticmethod
    def atmosphere_settings(identifier: str):
        return load_file("atmosphere_settings.jsont", {"format_version": PBR_SETTINGS_VERSION, "identifier": identifier}, is_json=True)

    @staticmethod
    def fog_settings(identifier: str):
        return load_file("fog_settings.jsont", {"format_version": FOG_VERSION, "identifier": identifier}, is_json=True)

    @staticmethod
    def shadow_settings():
        return load_file("shadow_settings.jsont", {"format_version": PBR_SETTINGS_VERSION}, is_json=True)

    @staticmethod
    def water_settings(identifier: str):
        return load_file("water_settings.jsont", {"format_version": PBR_SETTINGS_VERSION, "identifier": identifier}, is_json=True)

    @staticmethod
    def color_grading_settings(identifier: str):
        return load_file("color_grading_settings.jsont", {"format_version": PBR_SETTINGS_VERSION, "identifier": identifier}, is_json=True)

    @staticmethod
    def lighting_settings(identifier: str):
        return load_file("lighting_settings.jsont", {"format_version": PBR_SETTINGS_VERSION, "identifier": identifier}, is_json=True)

    @staticmethod
    def point_lights():
        return load_file("point_lights.jsont", {"format_version": PBR_SETTINGS_VERSION}, is_json=True)

    @staticmethod
    def pbr_fallback_settings():
        return load_file("pbr_fallback_settings.jsont", {"format_version": PBR_SETTINGS_VERSION}, is_json=True)

    @staticmethod
    def loot_table():
        return load_file("loot_table.jsont", is_json=True)

    @staticmethod
    def recipe_smelting(identifier: str, tags: list[str]):
        return load_file("recipe_smelting.jsont", {"format_version": RECIPE_JSON_FORMAT_VERSION, "identifier": identifier, "tags": tags}, is_json=True)

    @staticmethod
    def recipe_smithing_table(identifier: str, tags: list[str]):
        return load_file(
            "recipe_smithing_table.jsont", {"format_version": RECIPE_JSON_FORMAT_VERSION, "identifier": identifier, "tags": tags}, is_json=True
        )

    @staticmethod
    def recipe_smithing_table_trim(identifier: str, tags: list[str]):
        return load_file(
            "recipe_smithing_table_trim.jsont", {"format_version": RECIPE_JSON_FORMAT_VERSION, "identifier": identifier, "tags": tags}, is_json=True
        )

    @staticmethod
    def recipe_shapeless_crafting(identifier: str, tags: list[str]):
        return load_file(
            "recipe_shapeless_crafting.jsont", {"format_version": RECIPE_JSON_FORMAT_VERSION, "identifier": identifier, "tags": tags}, is_json=True
        )

    @staticmethod
    def recipe_shaped_crafting(identifier: str, assume_symmetry: bool, tags: list[str]):
        return load_file(
            "recipe_shaped_crafting.jsont",
            {"format_version": RECIPE_JSON_FORMAT_VERSION, "identifier": identifier, "tags": tags, "assume_symmetry": assume_symmetry},
            is_json=True,
        )
    
    @staticmethod
    def recipe_brewing_container(identifier: str, tags: list[str]):
        return load_file(
            "recipe_brewing_container.jsont", {"format_version": RECIPE_JSON_FORMAT_VERSION, "identifier": identifier, "tags": tags}, is_json=True
        )
    
    @staticmethod
    def recipe_brewing_mix(identifier: str, tags: list[str]):
        return load_file(
            "recipe_brewing_mix.jsont", {"format_version": RECIPE_JSON_FORMAT_VERSION, "identifier": identifier, "tags": tags}, is_json=True
        )

    @staticmethod
    def crafting_items_catalog():
        return load_file("crafting_items_catalog.jsont", {"format_version": CRAFTING_ITEMS_CATALOG}, is_json=True)

    @staticmethod
    def aim_assist_preset(identifier: str):
        return load_file("aim_assist_preset.jsont", {"identifier": identifier}, is_json=True)

    @staticmethod
    def aim_assist_categories():
        return load_file("aim_assist_categories.jsont", is_json=True)

    @staticmethod
    def jigsaw_structure_process(identifier: str):
        return load_file("jigsaw_structure_process.jsont", {"format_version": JIGSAW_VERSION, "identifier": identifier}, is_json=True)

    @staticmethod
    def jigsaw_structures(
        identifier: str,
        placement_step: str,
        start_pool_identifier: str,
        start_jigsaw_name: str | None,
        max_depth: int,
        start_height,
        liquid_settings,
    ):
        return load_file(
            "jigsaw_structures.jsont",
            {
                "format_version": JIGSAW_VERSION,
                "identifier": identifier,
                "placement_step": placement_step,
                "start_pool_identifier": start_pool_identifier,
                "start_jigsaw_name": start_jigsaw_name if start_jigsaw_name is not None else {},
                "max_depth": max_depth,
                "start_height": start_height,
                "liquid_settings": liquid_settings,
            },
            is_json=True,
        )

    @staticmethod
    def jigsaw_template_pools(identifier: str, fallback_identifier: str | None):
        return load_file(
            "jigsaw_template_pools.jsont",
            {
                "format_version": JIGSAW_VERSION,
                "identifier": identifier,
                "fallback_identifier": fallback_identifier if fallback_identifier else {},
            },
            is_json=True,
        )

    @staticmethod
    def jigsaw_structure_set(identifier: str, separation: int, spacing: int, spread_type, placement_type):
        return load_file(
            "jigsaw_structure_set.jsont",
            {
                "format_version": JIGSAW_VERSION,
                "identifier": identifier,
                "salt": salt_from_str(identifier),
                "separation": separation,
                "spacing": spacing,
                "spread_type": spread_type,
            },
            is_json=True,
        )

    @staticmethod
    def texture_set():
        return load_file(
            "texture_set.jsont",
            {
                "format_version": PBR_SETTINGS_VERSION,
            },
            is_json=True,
        )

    @staticmethod
    def python():
        return load_file("python.jsont")

    @staticmethod
    def gitignore():
        return load_file("gitignore.jsont")

    @staticmethod
    def code_workspace(name, path1, path2, preview=False):
        return load_file(
            "code_workspace.jsont",
            {
                "name": name,
                "path": os.path.join(path1, path2),
                "dev_res_path": os.path.join(
                    APPDATA,
                    "Local",
                    "Packages",
                    f"Microsoft.Minecraft{'WindowsBeta' if preview else 'UWP'}_8wekyb3d8bbwe",
                    "LocalState",
                    "games",
                    "com.mojang",
                    "development_resource_packs",
                ),
                "dev_beh_path": os.path.join(
                    APPDATA,
                    "Local",
                    "Packages",
                    f"Microsoft.Minecraft{'WindowsBeta' if preview else 'UWP'}_8wekyb3d8bbwe",
                    "LocalState",
                    "games",
                    "com.mojang",
                    "development_behavior_packs",
                ),
            },
            is_json=True,
        )

    @staticmethod
    def vscode(path, wkspc, script_uuid):
        return load_file("vscode.jsont", {"script_uuid": script_uuid, "wkspc": wkspc, "path": path}, is_json=True)

    @staticmethod
    def biome_server():
        return load_file("biome_server.jsont", {"format_version": BIOME_SERVER_VERSION}, is_json=True)

    @staticmethod
    def biome_client():
        return load_file("biome_client.jsont", {"format_version": BIOME_CLIENT_VERSION}, is_json=True)

    @staticmethod
    def block_culling_rules(identifier: Identifier):
        return load_file("block_culling.jsont", {"format_version": BLOCK_CULLING_VERSION, "identifier": identifier}, is_json=True)

class AddonDescriptor:
    """An object representing an addon descriptor with validation for names and namespaces."""

    _object_type = "addon_descriptor"

    def _validate_name(self, name: str, is_vanilla: bool = False, is_vanilla_allowed: bool = False) -> tuple[str, str]:
        """Validates the name of the Minecraft object.

        Parameters:
            name (str): The name of the Minecraft object.
        """
        object_name: str
        object_namespace: str

        if ":" in name:
            object_namespace, object_name = name.split(":", 1)
        else:
            object_namespace = "minecraft" if is_vanilla else CONFIG.NAMESPACE
            object_name = name

        if str(object_name)[0].isdigit():
            raise ValueError(f"Names cannot start with a digit. {object_name} at {self._object_type}[{name}]")

        if not str(object_namespace)[0].isalpha():
            raise ValueError(f"Namespaces cannot start with a digit. {self._object_type}[{name}]")

        if CONFIG._TARGET == ConfigPackageTarget.ADDON and is_vanilla and not is_vanilla_allowed:
            raise RuntimeError(f"Overriding vanilla features is not allowed for packages of type '{CONFIG._TARGET}'. {self._object_type}[{name}]")

        if is_vanilla and object_namespace != "minecraft":
            raise ValueError(f"Invalid namespace '{object_namespace}' overriding Vanilla for. {self._object_type}[{name}]")

        return object_name, object_namespace

    def __init__(self, name: str, is_vanilla: bool = False, is_vanilla_allowed: bool = False) -> None:
        """
        Constructs all the necessary attributes for the AddonObject object.

        Parameters:
            name (str): The name of the addon object.
        """

        self._is_vanilla = is_vanilla
        self._name, self._namespace = self._validate_name(str(name), is_vanilla, is_vanilla_allowed)
        self._display_name = self._name.replace("_", " ").title()
        self.__created_from = traceback.format_stack()[:-1]

    @property
    def identifier(self) -> Identifier:
        """
        Returns the identifier of the addon object in the format 'namespace:name'.

        Returns:
            str: The identifier of the addon object.
        """
        return f"{self._namespace}:{self._name}"

    @property
    def name(self) -> str:
        """
        Returns the name of the addon object.

        Returns:
            str: The name of the addon object.
        """
        return self._name

    def __str__(self):
        return self.identifier


# For the description property of json files
class MinecraftDescription(AddonDescriptor):
    """Handles Minecraft descriptions.

    Attributes:
        name (str): The name of the Minecraft object.
        is_vanilla (bool, optional): If the object is from vanilla Minecraft. Defaults to False.
    """

    def __init__(self, name, is_vanilla=False):
        super().__init__(name, is_vanilla)
        self._description: dict = JsonSchemes.description(self._namespace, self._name)

    def _export(self) -> dict:
        """Returns the description of the Minecraft object.

        Returns:
            dict: The description
        """
        return self._description


class AddonObject(AddonDescriptor):
    """
    An object representing an addon with functionality to modify its content, queue it for processing, and export it.

    Attributes:
        name (str): The name of the addon object.
        is_vanilla (bool): Indicates if the object is from vanilla Minecraft.
        object_type (str, optional): The type of the addon object.
    """

    _extension = ".json"
    _path = ""
    _object_type = "addon_object"

    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        """
        Constructs all the necessary attributes for the AddonObject object.

        Parameters:
            name (str): The name of the addon object.
        """
        super().__init__(name, is_vanilla)

        self._directory = ""
        self._content = {}
        self._shorten = True

    def do_not_shorten(self):
        """
        Setter property that disables shortening of `dict` when exporting.
        """
        self._shorten = False

    def content(self, content) -> AddonObject:
        """
        Sets the content of the addon object and returns the object.

        Parameters:
            content (any): The content to be set for the addon object.

        Returns:
            self: The instance of the current AddonObject.
        """
        self._content = content
        return self

    def queue(self, directory: str | None = None) -> AddonObject:
        """
        Queues the addon object for processing and logs the event.

        Parameters:
            directory (str, optional): The directory of the addon object.

        Returns:
            self: The instance of the current AddonObject.
        """
        from anvil import ANVIL

        self._directory = directory if not directory is None else ""
        self._path = os.path.join(self._path, self._directory)
        ANVIL._queue(self)
        return self

    def _export(self):
        """
        Exports the addon object after potentially shortening its content and replacing backslashes.
        Logs the event and writes the object to a file.
        """

        def _shorten_dict(d):
            if isinstance(d, dict):
                return {
                    k: (v if v != {"do_not_shorten": True} else {})
                    for k, v in ((k, _shorten_dict(v)) for k, v in d.items())
                    if (v != {} and v != [] and v != None) or str(k).startswith("minecraft:") or v == {"do_not_shorten": True}
                }

            elif isinstance(d, list):
                return [v for v in map(_shorten_dict, d) if v != []]

            return d

        def _replace_backslashes(obj):
            if isinstance(obj, str):
                return obj.replace('"/n"', '"\\n"').replace("/n", "\n").replace("\\", "/")
            elif isinstance(obj, list):
                return [_replace_backslashes(item) for item in obj]
            elif isinstance(obj, dict):
                return {key: _replace_backslashes(value) for key, value in obj.items()}
            else:
                return obj

        path = self._path.removeprefix(CONFIG.RP_PATH).removeprefix(CONFIG.BP_PATH)
        path = os.path.join(path, f"{self._name}{self._extension}")
        if len(path) > 80:
            raise ValueError(f"Relative file path [{path}] has [{len(path)}] characters, but cannot be more than [80] characters.")

        if self._shorten and type(self._content) is dict:
            self._content = _shorten_dict(self._content)

        self._content = _replace_backslashes(self._content)
        File(f"{self._name}{self._extension}", self._content, self._path, "w")


class MinecraftAddonObject(AddonDescriptor):
    def __init__(self, name, is_vanilla=False, is_vanilla_allowed=False):
        super().__init__(name, is_vanilla, is_vanilla_allowed)


class MinecraftEntityDescriptor(MinecraftAddonObject):
    _object_type = "Entity Descriptor"

    def __init__(self, name, is_vanilla=False, allow_runtime: bool = True, is_vanilla_allowed=True):
        super().__init__(name, is_vanilla, is_vanilla_allowed)
        self._allow_runtime = allow_runtime


class MinecraftBlockDescriptor(AddonDescriptor):
    _object_type = "Block Descriptor"

    def __init__(
        self,
        name,
        is_vanilla=False,
        states: Mapping[str, str | int | float | bool] = None,
        tags: set[str] = None,
        is_vanilla_allowed=True,
    ) -> None:
        super().__init__(name, is_vanilla, is_vanilla_allowed)
        self._states = {}
        if states:
            for k, v in states.items():
                if v is not None:
                    self._states[str(k)] = str(v)
        self._tags = tags if tags is not None else set()

    @property
    def tags(self) -> set[str]:
        """Returns the tags associated with the block."""
        return self._tags

    @property
    def states(self) -> str:
        """Returns a string representation of the block states."""
        if len(self._states.keys()) > 0:
            return "{" + ", ".join(f"{k}={v}" for k, v in self._states.items()) + "}"
        return ""

    def __str__(self) -> Identifier:
        if self.states:
            return f"{self.identifier} [{self.states}]"
        return self.identifier


class MinecraftItemDescriptor(AddonDescriptor):
    _object_type = "Item Descriptor"

    def __init__(self, name: str, is_vanilla: bool = False, is_vanilla_allowed: bool = True) -> None:
        super().__init__(name, is_vanilla, is_vanilla_allowed)

    def __str__(self) -> Identifier:
        return self.identifier


class MinecraftBiomeDescriptor(AddonDescriptor):
    _object_type = "Biome Descriptor"

    def __init__(self, name: str, is_vanilla: bool = False, is_vanilla_allowed: bool = True) -> None:
        super().__init__(name, is_vanilla, is_vanilla_allowed)

    def __str__(self) -> Identifier:
        return self.identifier


class EntityDescriptor(MinecraftEntityDescriptor):
    def __init__(self, name: str, is_vanilla: bool = False, allow_runtime: bool = True) -> None:
        super().__init__(name, is_vanilla, allow_runtime, False)


class BlockDescriptor(MinecraftBlockDescriptor):
    def __init__(self, name: str, is_vanilla: bool = False, states: Mapping[str, str | int | float | bool] = None, tags: set[str] = None) -> None:
        super().__init__(name, is_vanilla, states, tags, False)


class ItemDescriptor(MinecraftItemDescriptor):
    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        super().__init__(name, is_vanilla, False)


class BiomeDescriptor(MinecraftBiomeDescriptor):
    def __init__(self, name: str, is_vanilla: bool = False) -> None:
        super().__init__(name, is_vanilla, False)


class _BaseComponent(AddonDescriptor):
    _object_type = "Base Component"
    _target: ComponentTarget = ComponentTarget.Any

    def _require_components(self, *components: "_BaseComponent") -> None:
        self._dependencies.extend(components)

    def _add_clashes(self, *components: "_BaseComponent") -> None:
        self._clashes.extend(components)

    def _enforce_version(self, current: str, minimum: str) -> None:
        if Version(current) < Version(minimum):
            raise ValueError(f"{self.identifier} requires ≥ {minimum} (got {current}).")

    def _add_field(self, key: str, value: Any) -> None:
        self._component[key] = value

    def _set_value(self, value: Dict[str, Any]) -> None:
        self._component = value

    def _get_field(self, key: str, default: Any) -> Any:
        return self._component.get(key, default)

    def _add_dict(self, value: Dict[str, Any]) -> None:
        self._component.update(value)

    def __init__(self, component_name, is_vanilla: bool = True) -> None:
        super().__init__(component_name, is_vanilla, True)
        self._dependencies: List["_BaseComponent"] = []
        self._clashes: List["_BaseComponent"] = []
        self._component: Dict[str, Any] = {}

    def __iter__(self):
        """Iterates over the component's fields."""
        return iter({self.identifier: self._component}.items())
    
    def _export(self) -> Dict[str, Any]:
        """Exports the component as a dictionary.

        Returns:
            dict: The exported component.
        """
        return {self.identifier: self._component}



import os
from typing import Literal, TypedDict

from anvil import ANVIL, CONFIG
from anvil.lib.filters import Filter
from anvil.lib.lib import CopyFiles, FileExists, clamp
from anvil.lib.schemas import AddonObject, JsonSchemes
from anvil.lib.types import (ConstantIntProvider, Identifier,
                             StructureProcessors, UniformIntProvider)


class Structure:
    """A class representing a Structure."""

    def __init__(self, structure_name: str):
        """Initializes a Structure instance.

        Parameters:
            structure_name (str): The name of the structure.
        """
        splits = structure_name.split("/")
        self._sub_path = splits[:-1]
        self._name = splits[-1]
        self._path = os.path.join("structures", CONFIG.NAMESPACE, *self._sub_path, f"{self._name}.mcstructure")
        if not FileExists(os.path.join("world", "structures", *self._sub_path, f"{self._name}.mcstructure")):
            raise FileNotFoundError(
                f"{self._name}.mcstructure not found in {os.path.join('assets', 'world', 'structures')}. Please ensure the file exists."
            )

    def queue(self):
        """Queues the structure to be exported."""
        ANVIL._queue(self)

    @property
    def identifier(self) -> Identifier:
        """Returns the identifier of the structure."""
        return f"{CONFIG.NAMESPACE}:{self._name}"

    @property
    def reference(self) -> str:
        """Returns the reference path of the structure."""
        return self._path.removesuffix(".mcstructure").removeprefix("structures\\")

    def _export(self):
        """Exports the structure to the file system."""
        CopyFiles(
            os.path.join("world", "structures", *self._sub_path),
            os.path.join(CONFIG.BP_PATH, "structures", CONFIG.NAMESPACE, *self._sub_path),
            f"{self._name}.mcstructure",
        )


class _JigsawStructureProcess(AddonObject):
    """A class representing a Structure Process in Minecraft."""

    _extension = ".json"
    _path = os.path.join(CONFIG.BP_PATH, "worldgen", "processors")
    _object_type = "Jigsaw Structure Process"

    class predicates:
        class _input_predicate:
            def __init__(self, processor: dict):
                self._dict = processor

            def always_true(self):
                self._dict["input_predicate"] = {"predicate_type": "minecraft:always_true"}

            def block_match(self, identifier: str):
                self._dict["input_predicate"] = {"predicate_type": "minecraft:block_match", "block": identifier}

            def random_block_match(self, identifier: str, probability: float):
                self._dict["input_predicate"] = {
                    "predicate_type": "minecraft:random_block_match",
                    "block": identifier,
                    "probability": clamp(probability, 0.0, 1.0),
                }

            def block_state_match(self, identifier: str, states: dict[str, str | int | float | bool]):
                self._dict["input_predicate"] = {
                    "predicate_type": "minecraft:block_state_match",
                    "block_state": {"name": identifier, "states": {str(k): str(v) for k, v in states.items()}},
                }

            def random_block_state_match(self, identifier: str, probability: float, states: dict[str, str | int | float | bool]):
                self._dict["input_predicate"] = {
                    "predicate_type": "minecraft:random_block_state_match",
                    "block_state": {"name": identifier, "states": {str(k): str(v) for k, v in states.items()}},
                    "probability": clamp(probability, 0.0, 1.0),
                }

            def tag_match(self, tag: str):
                self._dict["input_predicate"] = {"predicate_type": "minecraft:tag_match", "tag": tag}

            def get(self):
                return self._dict

        class _position_predicate:
            def __init__(self, processor: dict):
                self._dict = processor

            def axis_aligned_linear_pos(
                self, min_chance: float, max_chance: float, min_distance: int, max_distance: int, axis: Literal["x", "y", "z"]
            ):
                self._dict["position_predicate"] = {
                    "predicate_type": "minecraft:axis_aligned_linear_pos",
                    "min_chance": clamp(min_chance, 0.0, 1.0),
                    "max_chance": clamp(max_chance, 0.0, 1.0),
                    "min_dist": min_distance,
                    "max_dist": max_distance,
                    "axis": axis,
                }

            def get(self):
                return self._dict

        class _location_predicate:
            def __init__(self, processor: dict):
                self._dict = processor

            def get(self):
                return self._dict

        def __init__(self, processor: dict):
            self.input_predicate = self._input_predicate(processor)
            self.position_predicate = self._position_predicate(processor)
            self.location_predicate = self._location_predicate(processor)

    def __init__(self, name: str):
        super().__init__(name)
        self.content(JsonSchemes.jigsaw_structure_process(self.identifier))

    def add_block_ignore_processor(self, blocks: list[str] | str):
        if isinstance(blocks, str):
            blocks = [blocks]
        if not isinstance(blocks, list):
            raise TypeError("blocks must be a list or a string")

        self._content["minecraft:processor_list"]["processors"].append({"processor_type": "minecraft:block_ignore", "blocks": blocks})

    def add_protected_blocks_processor(self, block_tag: str):
        if not isinstance(block_tag, str):
            raise TypeError("block_tag must be a string")

        self._content["minecraft:processor_list"]["processors"].append({"processor_type": "minecraft:protected_blocks", "value": block_tag})

    def add_capped_processor(self, delegate: StructureProcessors, limit: int | ConstantIntProvider | UniformIntProvider):
        if not isinstance(delegate, str):
            raise TypeError("delegate must be a string")
        if delegate not in ["minecraft:block_ignore", "minecraft:protected_blocks"]:
            raise ValueError("delegate must be one of 'minecraft:block_ignore', 'minecraft:protected_blocks'")
        if not isinstance(limit, (int, ConstantIntProvider, UniformIntProvider)):
            raise TypeError("limit must be an integer or a provider")

        if isinstance(limit, int):
            limit_data = limit
        elif isinstance(limit, ConstantIntProvider):
            limit_data = {"type": "constant", "value": limit["value"]}
        else:
            limit_data = {"type": "uniform", "min_inclusive": limit["min"], "max_inclusive": limit["max"]}

        self._content["minecraft:processor_list"]["processors"].append(
            {"processor_type": "minecraft:capped", "delegate": delegate, "limit": limit_data}
        )

    def add_block_rule(self, output_block: str, loot_table_path: str = None) -> predicates:
        if not isinstance(output_block, str):
            raise TypeError("output_block must be a string")
        if loot_table_path is not None and not isinstance(loot_table_path, str):
            raise TypeError("loot_table_path must be a string or None")

        processors: list[dict] = self._content["minecraft:processor_list"]["processors"]
        processor = next((pr for pr in processors if pr.get("processor_type") == "minecraft:rule"), None)
        if processor is None:
            processor = {"processor_type": "minecraft:rule", "rules": []}
            processors.append(processor)

        processor["rules"].append(
            {
                "input_predicate": {},
                "position_predicate": {},
                "location_predicate": {},
                "output_state": {"name": output_block},
            }
        )

        if loot_table_path:
            processor["block_entity_modifier"] = {
                "type": "minecraft:append_loot",
                "loot_table": loot_table_path,
            }

        predicate = _JigsawStructureProcess.predicates(processor["rules"][-1])
        return predicate

    def queue(
        self,
    ):
        return super().queue()


class _JigsawStructure(AddonObject):
    """A class representing a Jigsaw structure in Minecraft."""

    _extension = ".json"
    _path = os.path.join(CONFIG.BP_PATH, "worldgen", "jigsaw_structures")
    _object_type = "Jigsaw Structure"

    class _pool:
        def __init__(
            self,
        ):
            self._pools = []

        def direct_pool_alias(self, alias: str, target: str):
            """Creates a direct pool alias.

            Parameters:
                target (str): The target structure.
            """
            if not isinstance(target, str):
                raise TypeError("target must be a string")

            if not isinstance(alias, str):
                raise TypeError("alias must be a string")

            pool = _JigsawStructure._direct_pool_alias(alias, target)
            self._pools.append(pool)
            return pool

        def random_pool_aliases(self, alias: str):
            """Creates a random pool alias.

            Parameters:
                alias (str): The alias for the random pool.
            """
            if not isinstance(alias, str):
                raise TypeError("alias must be a string")

            pool = _JigsawStructure._random_pool_alias(alias)
            self._pools.append(pool)
            return pool

        def random_pool_group(self, weight: int):
            """Creates a random pool group.

            Parameters:
                weight (int): The weight of the random pool group.
            """
            if not isinstance(weight, int):
                raise TypeError("weight must be an integer")

            pool = _JigsawStructure._group_pool_alias()
            self._pools.append(pool)
            return pool

        def _export(self):
            """Exports the pool aliases to the content."""
            return {
                "pool_aliases": [pool._export() for pool in self._pools],
            }

    class _direct_pool_alias:
        def __init__(self, alias: str, target: str):
            """Initializes a direct pool alias.

            Parameters:
                alias (str): The alias for the direct pool.
                target (str): The target structure.
            """
            if not isinstance(alias, str):
                raise TypeError("alias must be a string")
            if not isinstance(target, str):
                raise TypeError("target must be a string")
            self._pool = {"type": "direct", "alias": alias, "target": target}

        def _export(self):
            """Exports the direct pool alias to the content."""
            return self._pool

    class _random_pool_alias:
        def __init__(self, alias: str):
            self._pool = {"type": "random", "alias": alias, "targets": []}

        def add_target(self, target: str, weight: int):
            """Adds a target to the random pool alias.

            Parameters:
                target (str): The target structure.
                weight (int): The weight of the target.
            """
            if not isinstance(target, str):
                raise TypeError("target must be a string")
            if not isinstance(weight, int):
                raise TypeError("weight must be an integer")
            self._pool["targets"].append({"data": target, "weight": weight})

        def _export(self):
            """Exports the random pool alias to the content."""
            return self._pool

    class _group_pool_alias(_pool):
        def __init__(self, weight: int):
            super().__init__()
            self._weight = weight

        def _export(self):
            return {
                "data": super()._export()["pool_aliases"],
                "weight": self._weight,
            }

    def __init__(
        self,
        name: str,
        start_pool: "JigsawStructureTemplatePool",
        max_depth: int,
        start_height: tuple[int, int] | int,
        placement_step: Literal[
            "raw_generation" "lakes",
            "local_modifications",
            "underground_structures",
            "surface_structures",
            "strongholds",
            "underground_ores",
            "underground_decoration",
            "fluid_springs",
            "vegetal_decoration",
            "top_layer_modification",
        ],
        start_jigsaw_name: str = None,
        liquid_settings: Literal["apply_waterlogging", "ignore_waterlogging"] = "apply_waterlogging",
        start_height_from_sea: bool = False,
    ):
        """Initializes a Jigsaw instance.

        Parameters:
            name (str): The name of the jigsaw structure.
            start_pool (JigsawStructureTemplatePool): The first Template Pool to use when generating the Jigsaw Structure.
            max_depth (int): The maximum recursion depth for Jigsaw Structure generation.
            start_height (tuple[int, int] | int): Height at which the Jigsaw Structure's start_pool should begin.
            placement_step (Literal[ 'raw_generation', 'lakes', 'local_modifications', 'underground_structures', 'surface_structures', 'strongholds', 'underground_ores', 'underground_decoration', 'fluid_springs', 'vegetal_decoration', 'top_layer_modification' ]): Specifies the world generation phase in which the Jigsaw Structure is generated.
            start_jigsaw_name (str, optional): The name of the Jigsaw Block from the start_pool to be placed first.
        """
        super().__init__(name)  # "jigsaw"

        if not isinstance(start_pool, JigsawStructureTemplatePool):
            raise TypeError("start_pool must be an instance of JigsawStructureTemplatePool")

        height = {}
        if isinstance(start_height, int):
            if start_height_from_sea:
                height = {"type": "constant", "value": {"from_sea": start_height}}
            else:
                height = {"type": "constant", "value": {"absolute": start_height}}

        elif isinstance(start_height, tuple) and len(start_height) == 2:
            height = {"type": "uniform", "min": {"above_bottom": min(start_height)}, "max": {"below_top": max(start_height)}}
        else:
            raise TypeError("start_height must be an int or a tuple of two ints")

        self.content(
            JsonSchemes.jigsaw_structures(
                self.identifier,
                placement_step,
                start_pool.identifier,
                start_jigsaw_name,
                clamp(max_depth, 0, 20),
                height,
                liquid_settings,
            )
        )
        self._pool_aliases: _JigsawStructure._pool = None
        self._start_pool = start_pool

    def add_biome_filters(self, filter: Filter):
        """Adds biome filters to the jigsaw structure.

        Parameters:
            *filters (Filter): The biome filters to add.
        """
        self._content["minecraft:jigsaw"]["biome_filters"].append(filter)

    def terrain_adaptation(self, terrain_adaptation: Literal["none", "bury", "beard_thin", "beard_box", "encapsulate"]):
        """Sets the terrain adaptation for the jigsaw structure.

        Parameters:
            terrain_adaptation (Literal['none', 'bury', 'beard_thin', 'beard_box', 'encapsulate']): The terrain adaptation to set.
            - "none" Do not adjust ambient block density.
            - "bury" Ambient block density will be added to all pieces of a structure, but only within the Y bounds of its starting piece. This is ideal for structures that need to bury themselves below the surface, but want another set of pieces to stick up through the terrain uncovered.
            - "beard_thin" Ambient block density will be added below the structure and block density will be reduced just above the ground.
            - "beard_box" Ambient block density will be added below the structure, and block density will be reduced within the entire box the structure occupies.
            - "encapsulate" Ambient block density will be added all around every piece of a structure.
        """
        if terrain_adaptation not in ["none", "bury", "beard_thin", "beard_box", "encapsulate"]:
            raise ValueError("Invalid terrain adaptation value")
        self._content["minecraft:jigsaw"]["terrain_adaptation"] = terrain_adaptation

    def heightmap_projection(self, heightmap_projection: Literal["none", "world_surface", "ocean_floor"]):
        """Sets the heightmap projection for the jigsaw structure.

        Parameters:
            heightmap_projection (Literal['none', 'world_surface', 'ocean_floor']): The heightmap projection to set.
        """
        if heightmap_projection not in ["none", "world_surface", "ocean_floor"]:
            raise ValueError("Invalid heightmap projection value")
        self._content["minecraft:jigsaw"]["heightmap_projection"] = heightmap_projection

    def dimension_padding(self, dimension_padding: tuple[int, int]):
        """Sets the dimension padding for the jigsaw structure.

        Parameters:
            dimension_padding (tuple[int, int]): The dimension padding to set.
        """
        if not isinstance(dimension_padding, tuple) or len(dimension_padding) != 2:
            raise TypeError("dimension_padding must be a tuple of two integers")
        self._content["minecraft:jigsaw"]["dimension_padding"] = dimension_padding

    def max_distance_from_center(self, horizontal: int, vertical: int):
        """Sets the maximum distance from center for the jigsaw structure.

        Parameters:
            horizontal (int): The maximum horizontal distance from center to set.
            vertical (int): The maximum vertical distance from center to set.
        """
        if not isinstance(horizontal, int):
            raise TypeError("horizontal must be an integer")
        if not isinstance(vertical, int):
            raise TypeError("vertical must be an integer")
        self._content["minecraft:jigsaw"]["max_distance_from_center"] = {
            "horizontal": horizontal,
            "vertical": vertical,
        }

    @property
    def pool_aliases(self) -> "_JigsawStructure._pool":
        """Returns the pool aliases for the jigsaw structure."""
        self._pool_aliases = self._pool()
        return self._pool_aliases

    def queue(self):
        return super().queue()

    def _export(self):
        self._content["minecraft:jigsaw"]["pool_aliases"] = self.pool_aliases._export()
        return super()._export()


class JigsawStructureTemplatePool(AddonObject):
    """A class representing a Structure Template Pool in Minecraft."""

    _extension = ".json"
    _path = os.path.join(CONFIG.BP_PATH, "worldgen", "template_pools")
    _object_type = "Jigsaw Structure Template Pool"

    def __init__(self, name: str, fallback: str = None):
        """Initializes a JigsawStructureTemplatePool instance.

        Parameters:
            name (str): The name of the template pool.
            fallback (str, optional): The fallback structure for the pool. Defaults to None.
        """
        super().__init__(name)  # "structure_template_pool"
        self.content(JsonSchemes.jigsaw_template_pools(self.identifier, fallback))
        self._structures: list[Structure] = []
        self._processors: list[_JigsawStructureProcess] = []

    def add_structure_element(
        self,
        structure: Structure | str | None,
        processors_name: str | _JigsawStructureProcess,
        weight: int = 1,
        projection: Literal["minecraft:rigid", "minecraft:terrain_matching"] = "minecraft:rigid",
    ) -> _JigsawStructureProcess | None:
        """Adds a structure to the template pool.

        Parameters:
            structure (str | Structure | None): The structure to add to the pool.
            weight (int, optional): The weight of the structure. Defaults to 1.
            processors_name (str, optional): The name of the processors for the structure. Defaults to None.
            projection (Literal["minecraft:rigid", "minecraft:terrain_matching"], optional): The projection type. Defaults to "minecraft:rigid".
        """
        if not isinstance(structure, (Structure, str, type(None))):
            raise TypeError("structure must be an instance of Structure or str")

        if not isinstance(processors_name, (str, _JigsawStructureProcess)):
            raise TypeError("processors_name must be an instance of str or JigsawStructureProcess")

        if isinstance(structure, str):
            structure = Structure(structure)
            self._structures.append(structure)

        if isinstance(processors_name, str):
            processors_name: _JigsawStructureProcess = _JigsawStructureProcess(processors_name)

        self._processors.append(processors_name)

        self._content["minecraft:template_pool"]["elements"].append(
            {
                "element": {
                    "element_type": "minecraft:single_pool_element" if structure else "minecraft:empty_pool_element",
                    "location": structure.reference if structure else {},
                    "processors": processors_name.identifier if processors_name else {},
                    # "terrain_projection": projection if projection else {},
                },
                "weight": weight,
            }
        )

        return processors_name

    def queue(self):
        for processor in self._processors:
            processor.queue()
        for structure in self._structures:
            structure.queue()
        super().queue()


class JigsawStructureSet(AddonObject):
    """A class representing a Structure Set in Minecraft.
    A Structure Set contains a set of Jigsaw Structures and rules for how those
    structures should be placed in the world relative to other instances of structures
    from the same set. Each structure within a set is paired with a weight that
    influences how frequently it is chosen.
    """

    _extension = ".json"
    _path = os.path.join(CONFIG.BP_PATH, "worldgen", "structure_sets")
    _structures: list[_JigsawStructure] = []
    _object_type = "Jigsaw Structure Set"

    def __init__(
        self,
        name: str,
        separation: int = 4,
        spacing: int = 10,
        spread_type: Literal["linear", "triangle"] = "linear",
        placement_type: Literal["minecraft:random_spread"] = "minecraft:random_spread",
    ):
        """Initializes a JigsawStructureSet instance.

        Parameters:
            name (str): The name of the structure set.
            separation (int): Padding (in chunks) within each grid cell. Structures will not generate within the padded area.
            spacing (int): Grid cell size (in chunks) to use when generating the structure. Structures will attempt to generate at a random position within each cell.
            spread_type (Literal["linear", "triangle"]): Randomness algorithm used when placing structures.
            placement_type (Literal["minecraft:random_spread"]): Describes where structures in the set spawn relative to one another. Currently, the only placement type supported is random_spread, which scatters structures randomly with a given separation and spacing.
        """
        super().__init__(name)  # "structure_set"
        self.content(JsonSchemes.jigsaw_structure_set(self.identifier, separation, spacing, spread_type, placement_type))

    def add_jigsaw_structure(
        self,
        structure_name: str,
        start_pool: "JigsawStructureTemplatePool",
        max_depth: int,
        start_height: tuple[int, int] | int,
        placement_step: Literal[
            "raw_generation" "lakes",
            "local_modifications",
            "underground_structures",
            "surface_structures",
            "strongholds",
            "underground_ores",
            "underground_decoration",
            "fluid_springs",
            "vegetal_decoration",
            "top_layer_modification",
        ],
        weight: int = 1,
        start_jigsaw_name: str = None,
        liquid_settings: Literal["apply_waterlogging", "ignore_waterlogging"] = "apply_waterlogging",
        start_height_from_sea: bool = False,
    ):
        """Adds a jigsaw structure to the structure set.

        Parameters:
            structure_name (str): The name of the jigsaw structure to add.
            weight (int): The weight of the structure in the set. Defaults to 1.
        """

        structure = _JigsawStructure(
            structure_name, start_pool, max_depth, start_height, placement_step, start_jigsaw_name, liquid_settings, start_height_from_sea
        )
        self._structures.append(structure)
        self._content["minecraft:structure_set"]["structures"].append({"structure": structure.identifier, "weight": weight})
        return structure

    def queue(self):
        for structure in self._structures:
            structure.queue()

        return super().queue()



import os

from anvil import ANVIL, CONFIG
from anvil.api.logic.molang import Molang
from anvil.api.pbr.pbr import __TextureSet
from anvil.lib.blockbench import _Blockbench
from anvil.lib.enums import (BlockFaces, BlockLiquidDetectionTouching,
                             BlockMaterial, BlockMovementType, TintMethod)
from anvil.lib.format_versions import (BLOCK_JSON_FORMAT_VERSION,
                                       BLOCK_SERVER_VERSION,
                                       ITEM_SERVER_VERSION)
from anvil.lib.lib import CopyFiles, FileExists, clamp
from anvil.lib.schemas import _BaseComponent
from anvil.lib.types import RGB, RGBA


class BlockDestructionParticles(_BaseComponent):
    _identifier = "minecraft:destruction_particles"

    def __init__(self, blockbench_name: str, texture: str, particle_count: int, tint_method: TintMethod = TintMethod.None_) -> None:
        """Sets the particles that will be used when the block is destroyed.

        Parameters:
            blockbench_name (str): The name of the blockbench model.
            texture (str): The texture name used for the particle.
            particle_count (int): The number of particles to spawn on destruction.
            tint_method (TintMethod, optional): Tint multiplied to the color. Defaults to TintMethod.None_.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blockcomponents/minecraft_destruction_particles
        """
        super().__init__("destruction_particles")
        self._enforce_version(BLOCK_SERVER_VERSION, "1.21.100")
        bb = _Blockbench(blockbench_name, "blocks")
        bb.textures.queue_texture(texture)
        self._add_field("texture", texture)

        self._add_field("particle_count", particle_count)

        if tint_method is not TintMethod.None_:
            self._add_field("tint_method", tint_method)


class BlockMovable(_BaseComponent):
    _identifier = "minecraft:movable"

    def __init__(self, movement_type: BlockMovementType = BlockMovementType.PushPull, sticky: str = "none") -> None:
        """The description identifier of the movable component.

        Parameters:
            movement_type (BlockMovementType, optional): How the block reacts to being pushed by another block like a piston. Defaults to BlockMovementType.PushPull.
            sticky (str, optional): How the block should handle adjacent blocks around it when being pushed by another block like a piston. Defaults to "none".

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blockcomponents/minecraft_movable
        """
        super().__init__("movable")
        self._enforce_version(BLOCK_SERVER_VERSION, "1.21.100")
        self._add_field("movement_type", movement_type)
        self._add_field("sticky", sticky)


class BlockRedstoneConductivity(_BaseComponent):
    _identifier = "minecraft:redstone_conductivity"

    def __init__(self, allows_wire_to_step_down: bool = True, redstone_conductor: bool = False) -> None:
        """Specifies whether a block has redstone properties. If the component is not provided, the default values are used.

        Parameters:
            allows_wire_to_step_down (bool, optional): Specifies if redstone wire can stair-step downward on the block, Defaults to True.
            redstone_conductor (bool, optional): Specifies if the block can be powered by redstone, Defaults to False.

        [Documentation reference]: https://learn.microsoft.com/en-gb/minecraft/creator/reference/content/blockreference/examples/blockcomponents/minecraft_redstone_conductivity
        """
        super().__init__("redstone_conductivity")
        self._enforce_version(BLOCK_SERVER_VERSION, "1.21.30")
        self._add_field("allows_wire_to_step_down", allows_wire_to_step_down)
        self._add_field("redstone_conductor", redstone_conductor)


class BlockCustomComponents(_BaseComponent):
    _identifier = "minecraft:custom_components"

    def __init__(self, component_name: str) -> None:
        """Allows you to add custom components to a block by specifying the components in a comma-separated list.

        Parameters:
            components (str): The components to register, if the namespace is not provided, the project namespace will be used.

        [Documentation reference]: https://learn.microsoft.com/en-us/minecraft/creator/documents/scripting/custom-components
        """
        super().__init__(component_name, False)
        self._set_value({"do_not_shorten": True})


class BlockDefault(_BaseComponent):
    _identifier = "minecraft:block_default"

    def __init__(self) -> None:
        """The default block component."""
        super().__init__("block_default")

    def ambient_occlusion_exponent(self, exponent: int):
        """The exponent for ambient occlusion of the block."""
        self._add_field("ambient_occlusion_exponent", max(0, exponent))
        return self

    def sound(self, sound: str):
        """The sound of the block."""
        self._add_field("sound", sound)
        return self

    def isotropic(self, **kwParameters: dict[str, BlockFaces]):
        """The isotropic of the block."""
        if BlockFaces.All in kwParameters.values():
            self._set_value("isotropic", True)
        else:
            self._add_field("isotropic", {v: k for k, v in kwParameters.items()})
        return self

    def textures(self, **kwParameters: dict[str, BlockFaces]):
        """The textures of the block."""
        for k in kwParameters.keys():
            if not FileExists(os.path.join("assets", "textures", "blocks", f"{k}.png")):
                raise FileNotFoundError(
                    f"{k}.png not found in {os.path.join("assets", "textures", "blocks")}. Please ensure the file exists."
                )

            CopyFiles(
                os.path.join("assets", "textures", "blocks"),
                os.path.join(CONFIG.RP_PATH, "textures", CONFIG.NAMESPACE, CONFIG.PROJECT_NAME, "blocks").replace("\\", "/"),
                f"{k}.png",
            )
            ANVIL.definitions.register_terrain_texture(k, "", k)

        if BlockFaces.All in kwParameters.values():
            self._add_field("textures", [f"{CONFIG.NAMESPACE}:{k}" for k, v in kwParameters.items() if v == BlockFaces.All][0])
        else:
            self._add_field("textures", {v: f"{CONFIG.NAMESPACE}:{k}" for k, v in kwParameters.items()})

        return self

    def carried_textures(self, **kwParameters: dict[str, BlockFaces]):
        """The carried textures of the block."""
        if BlockFaces.All in kwParameters.values():
            self._set_value("carried_textures", [f"{CONFIG.NAMESPACE}:{k}" for k, v in kwParameters.items() if v == BlockFaces.All][0])
        else:
            self._add_field("carried_textures", {v: f"{CONFIG.NAMESPACE}:{k}" for k, v in kwParameters.items()})
        return self


class BlockDestructibleByExplosion(_BaseComponent):
    _identifier = "minecraft:destructible_by_explosion"

    def __init__(self, explosion_resistance: int = None) -> None:
        """Describes the destructible by explosion properties for this block.

        Parameters:
            explosion_resistance (int): The amount of resistance to explosions in a range of 0 to 100.
        """
        super().__init__("destructible_by_explosion")

        if explosion_resistance is None:
            self._set_value(False)
        else:
            self._add_field("explosion_resistance", explosion_resistance)


class BlockDestructibleByMining(_BaseComponent):
    _identifier = "minecraft:destructible_by_mining"

    def __init__(self, seconds_to_destroy: int = None) -> None:
        """Describes the destructible by mining properties for this block.

        Parameters:
            seconds_to_destroy (int, optional): The amount of time it takes to destroy the block in seconds. If None, disables mining destructibility.
        """
        super().__init__("destructible_by_mining")
        self._enforce_version(ITEM_SERVER_VERSION, "1.21.50")
        if seconds_to_destroy is None:
            self._set_value(False)
        else:
            self._add_field("seconds_to_destroy", seconds_to_destroy)
            self._add_field("item_specific_speeds", [])

    def item_specific_speeds_name(self, destroy_speed: float, item_name: str):
        if "item_specific_speeds" in self._component:
            self._component["item_specific_speeds"].append({"item": item_name, "destroy_speed": destroy_speed})

    def item_specific_speeds_tag(self, destroy_speed: float, item_tag: str | Molang):
        if "item_specific_speeds" in self._component:
            self._component["item_specific_speeds"].append({"item": {"tags": item_tag}, "destroy_speed": destroy_speed})


class BlockFlammable(_BaseComponent):
    _identifier = "minecraft:flammable"

    def __init__(self, catch_chance_modifier: int, destroy_chance_modifier: int) -> None:
        """Describes the flammable properties for this block.

        Parameters:
            catch_chance_modifier (int): The chance that this block will catch fire in a range of 0 to 100.
            destroy_chance_modifier (int): The chance that this block will be destroyed when on fire in a range of 0 to 100.
        """
        super().__init__("flammable")
        self._add_field("catch_chance_modifier", catch_chance_modifier)
        self._add_field("destroy_chance_modifier", destroy_chance_modifier)


class BlockFriction(_BaseComponent):
    _identifier = "minecraft:friction"

    def __init__(self, friction: float = 0.4) -> None:
        """Describes the friction for this block in a range of 0.0 to 0.9.

        Parameters:
            friction (float, optional): The friction of the block. Defaults to 0.4.
        """
        super().__init__("flammable")
        self._set_value(clamp(friction, 0, 0.9))


class BlockLightDampening(_BaseComponent):
    _identifier = "minecraft:light_dampening"

    def __init__(self, light_dampening: int = 15) -> None:
        """The amount that light will be dampened when it passes through the block in a range of 0 to 15.

        Parameters:
            light_dampening (int, optional): The amount of light dampening. Defaults to 15.
        """
        super().__init__("light_dampening")
        self._set_value(clamp(light_dampening, 0, 15))


class BlockLightEmission(_BaseComponent):
    _identifier = "minecraft:light_emission"

    def __init__(self, light_emission: int = 0) -> None:
        """The amount of light this block will emit in a range of 0 to 15.

        Parameters:
            light_emission (int, optional): The amount of light emission. Defaults to 0.
        """
        super().__init__("light_emission")
        self._set_value(clamp(light_emission, 0, 15))


class BlockLootTable(_BaseComponent):
    _identifier = "minecraft:loot"

    def __init__(self, loot: str) -> None:
        """Specifies the path to the loot table.

        Parameters:
            loot (str): The path to the loot table.
        """
        super().__init__("loot")
        self._set_value(loot)


class BlockMapColor(_BaseComponent):
    _identifier = "minecraft:map_color"

    def __init__(self, color: str, tint_method: TintMethod = TintMethod.None_) -> None:
        """Sets the color of the block when rendered to a map. If this component is omitted, the block will not show up on the map.

        Parameters:
            color (str): The color is represented as a hex value in the format "#RRGGBB". May also be expressed as an array of [R, G, B] from 0 to 255.
            tint_method (str, optional): Optional, tint multiplied to the color. Tint method logic varies, but often refers to the "rain" and "temperature" of the biome the block is placed in to compute the tint. Defaults to None.

        [Documentation reference]: https://learn.microsoft.com/en-us/minecraft/creator/reference/content/blockreference/examples/blockcomponents/minecraftblock_map_color
        """
        self._enforce_version(BLOCK_JSON_FORMAT_VERSION, "1.21.70")
        super().__init__("map_color")
        self._add_field("color", color)
        if tint_method is not TintMethod.None_:
            self._add_field("tint_method", tint_method)


class BlockMaterialInstance(_BaseComponent):
    _identifier = "minecraft:material_instances"

    def __init__(self) -> None:
        """Maps face or material_instance names in a geometry file to an actual material instance."""
        super().__init__("material_instances")
        self._texture_set: __TextureSet = None

    def add_instance(
        self,
        blockbench_name: str,
        color_texture: str,
        block_face: BlockFaces = BlockFaces.All,
        render_method: BlockMaterial = BlockMaterial.Opaque,
        ambient_occlusion: float = 0,
        face_dimming: bool = True,
        isotropic: bool = False,
        tint_method: TintMethod = TintMethod.None_,
        normal_texture: str | RGB | RGBA = None,
        heightmap_texture: str | RGB | RGBA = None,
        metalness_emissive_roughness_texture: str | RGB | RGBA = None,
        metalness_emissive_roughness_subsurface_texture: str | RGB | RGBA = None,
    ):
        """Maps face or material_instance names in a geometry file to an actual material instance.

        Parameters:
            color_texture (str): The name of the texture to use for this block.
            block_face (BlockFaces, optional): The face of the block to apply the texture to. Defaults to BlockFaces.All.
            render_method (BlockMaterial, optional): The render method to use for this block. Defaults to BlockMaterial.Opaque.
            ambient_occlusion (float, optional): The amount of ambient occlusion for this block. Defaults to 0.
            face_dimming (bool, optional): Whether or not to use face dimming for this block. Defaults to True.
            tint_method (TintMethod, optional): The tint method to use for this block. Defaults to TintMethod.None_.


        """
        bb = _Blockbench(blockbench_name, "blocks")
        bb.textures.queue_texture(color_texture)

        ANVIL.definitions.register_terrain_texture(color_texture, blockbench_name, color_texture)

        self._add_dict(
            {
                "*" if block_face == BlockFaces.All else block_face: {
                    "texture": f"{CONFIG.NAMESPACE}:{color_texture}",
                    "render_method": render_method if not render_method == BlockMaterial.Opaque else {},
                    "ambient_occlusion": ambient_occlusion if ambient_occlusion is False else {},
                    "face_dimming": face_dimming if face_dimming is False else {},
                    "tint_method": tint_method if not tint_method == TintMethod.None_ else {},
                    "isotropic": isotropic if isotropic is True else {},
                }
            }
        )

        if any(
            [
                normal_texture,
                heightmap_texture,
                metalness_emissive_roughness_texture,
                metalness_emissive_roughness_subsurface_texture,
            ]
        ):
            self._texture_set = __TextureSet(self.identifier, "blocks")
            self._texture_set.set_textures(
                blockbench_name,
                color_texture,
                normal_texture,
                heightmap_texture,
                metalness_emissive_roughness_texture,
                metalness_emissive_roughness_subsurface_texture,
            )

        return self

    def __getitem__(self, key):
        if self._texture_set:
            self._texture_set.queue()
        return super().__getitem__(key)


class BlockGeometry(_BaseComponent):
    _identifier = "minecraft:geometry"

    def __init__(self, geometry_name: str, uv_lock: bool = False) -> None:
        """The description identifier of the geometry file to use to render this block.

        Parameters:
            geometry_name (str): The name of the geometry to use to render this block.
            uv_lock (bool, optional): A Boolean locking UV orientation of all bones in the geometry, or an array of strings locking UV orientation of specific bones in the geometry. For performance reasons it is recommended to use the Boolean. Note that for cubes using Box UVs, rather than Per-face UVs, 'uv_lock' is only supported if the cube faces are square.
        """
        super().__init__("geometry")
        self._enforce_version(BLOCK_SERVER_VERSION, "1.21.80")
        self._add_field("identifier", f"geometry.{CONFIG.NAMESPACE}.{geometry_name}")
        self._geometry_name = geometry_name
        if uv_lock:
            self._add_field("uv_lock", uv_lock)

        self._bb = _Blockbench(geometry_name, "blocks")
        self._bb.model.queue_model()

    def bone_visibility(self, **bone: dict[str, bool | str | Molang]):
        """Specifies the visibility of bones in the geometry file.

        Example:
            >>> BlockGeometry('block').bone_visibility(bone0=True, bone1=False)

        """
        self._add_field("bone_visibility", {b: v for b, v in bone.items()})
        return self

    def block_culling(self):
        self._add_field("culling", f"{CONFIG.NAMESPACE}:{self._geometry_name}")
        return self._bb.model.block_culling()


class BlockCollisionBox(_BaseComponent):
    _identifier = "minecraft:collision_box"

    def __init__(self, size: list[float], origin: list[float]) -> None:
        """Defines the area of the block that collides with entities.

        Parameters:
            size (list[float]): The size of the collision box.
            origin (list[float]): The origin of the collision box.
        """
        super().__init__("collision_box")
        min_origin = (-8, 0, -8)
        max_origin = (8, 16, 8)
        if size == (0, 0, 0):
            self._set_value(False)
        else:
            if any(o < min_o or o > max_o for o, min_o, max_o in zip(origin, min_origin, max_origin)):
                raise ValueError(f"Origin {origin} must be within {min_origin} and {max_origin}.")
            self._add_field("size", size)
            self._add_field("origin", origin)


class BlockSelectionBox(_BaseComponent):
    _identifier = "minecraft:selection_box"

    def __init__(self, size: list[float], origin: list[float]) -> None:
        """Defines the area of the block that is selected by the player's cursor.

        Parameters:
            size (list[float]): The size of the selection box.
            origin (list[float]): The origin of the selection box.
        """
        super().__init__("selection_box")
        if size == (0, 0, 0):
            self._set_value(False)
        else:
            self._add_field("size", size)
            self._add_field("origin", origin)


class BlockPlacementFilter(_BaseComponent):
    _identifier = "minecraft:placement_filter"

    def __init__(self) -> None:
        """By default, custom blocks can be placed anywhere and do not have placement restrictions unless you specify them in this component."""
        super().__init__("placement_filter")
        self._add_field("conditions", [])

    def add_condition(self, allowed_faces: list[BlockFaces], block_filter: list[str]):
        """Adds a condition to the placement filter.

        Parameters:
            allowed_faces (list[BlockFaces]): The faces of the block that are allowed to be placed on.
            block_filter (list[BlockDescriptor | str]): The blocks that are allowed to be placed on.
        """
        if BlockFaces.Side in allowed_faces:
            allowed_faces.remove(BlockFaces.North)
            allowed_faces.remove(BlockFaces.South)
            allowed_faces.remove(BlockFaces.East)
            allowed_faces.remove(BlockFaces.West)
        if BlockFaces.All in allowed_faces:
            allowed_faces = [BlockFaces.All]
        self._component["conditions"].append(
            {
                "allowed_faces": allowed_faces,
                "block_filter": block_filter,
            }
        )
        return self


class BlockTransformation(_BaseComponent):
    _identifier = "minecraft:transformation"

    def __init__(self) -> None:
        """The block's transfomration."""
        super().__init__("transformation")

    def translation(self, translation: tuple[float, float, float]):
        """The block's translation.

        Parameters:
            translation (position): The block's translation.
        """
        self._add_field("translation", translation)
        return self

    def scale(self, scale: tuple[float, float, float], scale_pivot: tuple[float, float, float] = (0, 0, 0)):
        """The block's scale.

        Parameters:
            scale (position): The block's scale.
        """
        self._add_field("scale", scale)

        if scale_pivot != (0, 0, 0):
            self._add_field("scale_pivot", scale_pivot)

        return self

    def rotation(self, rotation: tuple[float, float, float], rotation_pivot: tuple[float, float, float] = (0, 0, 0)):
        """The block's rotation.

        Parameters:
            rotation (position): The block's rotation.

        """
        self._add_field("rotation", rotation)

        if rotation_pivot != (0, 0, 0):
            self._add_field("rotation_pivot", rotation_pivot)
        return self


class BlockDisplayName(_BaseComponent):
    _identifier = "minecraft:display_name"

    def __init__(self, display_name: str, localize: bool = True) -> None:
        """Sets the block display name within Minecraft: Bedrock Edition. This component may also be used to pull from the localization file by referencing a key from it.

        Parameters:
            display_name (str): Set the display name for an block.
            localize (bool, optional): Whether to use the name with a localization file or not. Defaults to True.

        """
        super().__init__("display_name")
        if localize:
            key = f'tile.{CONFIG.NAMESPACE}:{display_name.lower().replace(" ", "_")}.name'
            ANVIL.definitions.register_lang(key, display_name)
            self._set_value(key)
        else:
            self._set_value(display_name)


class BlockCraftingTable(_BaseComponent):
    _identifier = "minecraft:crafting_table"

    def __init__(self, table_name: str, *crafting_tags: str) -> None:
        """Makes your block into a custom crafting table which enables the crafting table UI and the ability to craft recipes.

        Parameters:
            table_name (str): The name of the crafting table.

        Raises:
            IndexError: The crafting table tags cannot exceed 64 tags.
            ValueError: The crafting table tags are limited to 64 characters.
        """
        super().__init__("crafting_table")
        self._add_field("table_name", table_name)

        if len(crafting_tags) > 64:
            raise IndexError("Crafting Table tags cannot exceed 64 tags.")

        for tag in crafting_tags:
            if len(tag) > 64:
                raise ValueError("Crafting Table tags are limited to 64 characters.")

        self._add_field("crafting_tags", crafting_tags)


class BlockItemVisual(_BaseComponent):
    _identifier = "minecraft:item_visual"

    def __init__(self, geometry_name: str, texture: str, render_method: BlockMaterial = BlockMaterial.Opaque) -> None:
        """The description identifier of the geometry and material used to render the item of this block.

        Parameters:
            geometry (str): The geometry of the item.
            texture (str): The texture of the item.
            render_method (BlockMaterial): The method used to render the item.

        """
        self._enforce_version(BLOCK_SERVER_VERSION, "1.21.60")
        super().__init__("item_visual")

        bb = _Blockbench(geometry_name, "blocks")
        bb.model.queue_model()
        bb.textures.queue_texture(texture)

        ANVIL.definitions.register_terrain_texture(texture, geometry_name, texture)

        self._add_field("geometry", {"identifier": f"geometry.{CONFIG.NAMESPACE}.{geometry_name}"})
        self._add_field("material_instances", {"*": {"texture": f"{CONFIG.NAMESPACE}:{texture}", "render_method": render_method}})


class BlockLiquidDetection(_BaseComponent):
    _identifier = "minecraft:liquid_detection"

    def __init__(self) -> None:
        """The block's liquid detection."""
        self._enforce_version(BLOCK_SERVER_VERSION, "1.21.60")
        super().__init__("liquid_detection")
        self._add_field("detection_rules", [])

    def add_rule(
        self,
        liquid_type: str = "minecraft:water",
        on_liquid_touches: BlockLiquidDetectionTouching = BlockLiquidDetectionTouching.Blocking,
        can_contain_liquid: bool = False,
        stops_liquid_flowing_from_direction: list[BlockFaces] = [],
    ):
        """Adds a rule to the liquid detection.

        Parameters:
            liquid_type (str): The type of liquid, defaults to "minecraft:water".
            on_liquid_touches (BlockLiquidDetectionTouching): The action to take when the liquid touches the block.
            can_contain_liquid (bool, optional): Whether the block can contain the liquid. Defaults to False.

        """
        self._component["detection_rules"].append(
            {
                "liquid_type": "minecraft:water",
                "on_liquid_touches": on_liquid_touches.value,
                "can_contain_liquid": can_contain_liquid,
                "stops_liquid_flowing_from_direction": (
                    [BlockFaces.North, BlockFaces.South, BlockFaces.East, BlockFaces.West]
                    if BlockFaces.Side in stops_liquid_flowing_from_direction
                    else (
                        [BlockFaces.North, BlockFaces.South, BlockFaces.East, BlockFaces.West, BlockFaces.Up, BlockFaces.Down]
                        if BlockFaces.All in stops_liquid_flowing_from_direction
                        else stops_liquid_flowing_from_direction
                    )
                ),
            }
        )
        return self


class BlockDestructionParticles(_BaseComponent):
    _identifier = "minecraft:destruction_particles"

    def __init__(self, blockbench_name: str, texture: str = None, tint_method: TintMethod = TintMethod.None_) -> None:
        """Sets the particles that will be used when the block is destroyed.

        Parameters:
            blockbench_name (str): The name of the blockbench model.
            texture (str, optional): The texture name used for the particle.
            tint_method (TintMethod, optional): Tint multiplied to the color. Defaults to TintMethod.None_.
        """
        super().__init__("destruction_particles")
        if texture is not None:
            bb = _Blockbench(blockbench_name, "blocks")
            bb.textures.queue_texture(texture)
            self._add_field("texture", texture)
        if tint_method is not TintMethod.None_:
            self._add_field("tint_method", tint_method)


class BlockTick(_BaseComponent):
    _identifier = "minecraft:tick"

    def __init__(
        self,
        interval_range: tuple[int, int] = (0, 0),
        looping: bool = True,
    ) -> None:
        """Causes the block to tick at intervals randomly chosen from interval_range.

        Parameters:
            interval_range (tuple[int, int], optional): [min, max] ticks before next tick. Defaults to (0, 0).
            looping (bool, optional): If True, block will continue ticking; otherwise it ticks only once. Defaults to True.

        [Documentation reference]: https://learn.microsoft.com/en-us/minecraft/creator/reference/content/blockreference/examples/blockcomponents/minecraftblock_tick
        """
        super().__init__("tick")
        # ensure first value <= second value
        low, high = interval_range
        if low > high:
            low, high = high, low
        self._add_field("interval_range", [low, high])
        self._add_field("looping", looping)
